# The MIT License (MIT)
#
# Copyright (c) 2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re
import typing
from typing import Optional, List, Iterable
from barectf.typing import Index, _OptStr


__all__ = ['OptDescr', '_OptItem', '_NonOptItem', '_Error', 'parse', 'OrigArgs']


# types
OrigArgs = List[str]


# Option descriptor.
class OptDescr:
    # Builds an option descriptor having the short name `short_name`
    # (without the leading `-`) and/or the long name `long_name`
    # (without the leading `--`).
    #
    # If `has_arg` is `True`, then it is expected that such an option
    # has an argument.
    def __init__(self, short_name: _OptStr = None, long_name: _OptStr = None,
                 has_arg: bool = False):
        assert short_name is not None or long_name is not None
        self._short_name = short_name
        self._long_name = long_name
        self._has_arg = has_arg

    @property
    def short_name(self) -> _OptStr:
        return self._short_name

    @property
    def long_name(self) -> _OptStr:
        return self._long_name

    @property
    def has_arg(self) -> Optional[bool]:
        return self._has_arg


class _Item:
    pass


# Parsed option argument item.
class _OptItem(_Item):
    def __init__(self, descr: OptDescr, arg_text: _OptStr = None):
        self._descr = descr
        self._arg_text = arg_text

    @property
    def descr(self) -> OptDescr:
        return self._descr

    @property
    def arg_text(self) -> _OptStr:
        return self._arg_text


# Parsed non-option argument item.
class _NonOptItem(_Item):
    def __init__(self, text: str, orig_arg_index: Index, non_opt_index: Index):
        self._text = text
        self._orig_arg_index = orig_arg_index
        self._non_opt_index = non_opt_index

    @property
    def text(self) -> str:
        return self._text

    @property
    def orig_arg_index(self) -> Index:
        return self._orig_arg_index

    @property
    def non_opt_index(self) -> Index:
        return self._non_opt_index


# Results of parse().
class _ParseRes:
    def __init__(self, items: List[_Item], ingested_orig_args: OrigArgs,
                 remaining_orig_args: OrigArgs):
        self._items = items
        self._ingested_orig_args = ingested_orig_args
        self._remaining_orig_args = remaining_orig_args

    @property
    def items(self) -> List[_Item]:
        return self._items

    @property
    def ingested_orig_args(self) -> OrigArgs:
        return self._ingested_orig_args

    @property
    def remaining_orig_args(self) -> OrigArgs:
        return self._remaining_orig_args


# Parsing error.
class _Error(Exception):
    def __init__(self, orig_arg_index: Index, orig_arg: str, msg: str):
        super().__init__(msg)
        self._orig_arg_index = orig_arg_index
        self._orig_arg = orig_arg
        self._msg = msg

    @property
    def orig_arg_index(self) -> Index:
        return self._orig_arg_index

    @property
    def orig_arg(self) -> str:
        return self._orig_arg

    @property
    def msg(self) -> str:
        return self._msg


# Results of parse_short_opts() and parse_long_opt(); internal.
class _OptParseRes(typing.NamedTuple):
    items: List[_Item]
    orig_arg_index_incr: int


# Parses the original arguments `orig_args` (list of strings),
# considering the option descriptors `opt_descrs` (set of `OptDescr`
# objects), and returns a corresponding `_ParseRes` object.
#
# This function considers ALL the elements of `orig_args`, including the
# first one, so that you would typically pass `sys.argv[1:]` to exclude
# the program/script name.
#
# This argument parser supports:
#
# * Short options without an argument, possibly tied together:
#
#       -f -auf -n
#
# * Short options with arguments:
#
#       -b 45 -f/mein/file -xyzhello
#
# * Long options without an argument:
#
#       --five-guys --burger-king --pizza-hut --subway
#
# * Long options with arguments:
#
#       --security enable --time=18.56
#
# * Non-option arguments (anything else).
#
# This function does NOT accept `--` as an original argument; while it
# means "end of options" for many command-line tools, this function is
# all about keeping the order of the arguments, so it doesn't mean much
# to put them at the end. This has the side effect that a non-option
# argument cannot have the form of an option, for example if you need to
# pass the exact relative path `--lentil-soup`. In that case, you would
# need to pass `./--lentil-soup`.
#
# This function accepts duplicate options (the resulting list of items
# contains one entry for each instance).
#
# On success, this function returns a `_ParseRes` object which contains
# a list of items as its `items` property. Each item is either an
# option item or a non-option item.
#
# The returned list contains the items in the same order that the
# original arguments `orig_args` were parsed, including non-option
# arguments. This means, for example, that for
#
#     --hello --meow=23 /path/to/file -b
#
# the function creates a list of four items: two options, one
# non-option, and one option.
#
# In the returned object, `ingested_orig_args` is the list of ingested
# original arguments to produce the resulting items, while `remaining_orig_args`
# is the list of remaining original arguments (not parsed because an
# unknown option was found and `fail_on_unknown_opt` was `False`).
#
# For example, with
#
#     --great --white contact nuance --shark nuclear
#
# if `--shark` is not described within `opt_descrs` and
# `fail_on_unknown_opt` is `False`, then `ingested_orig_args` contains
# `--great`, `--white`, `contact`, and `nuance` (two options, two
# non-options), whereas `remaining_orig_args` contains `--shark` and
# `nuclear`.
#
# This makes it possible to know where a command name is, for example.
# With those arguments:
#
#     --verbose --stuff=23 do-something --specific-opt -f -b
#
# and the option descriptors for `--verbose` and `--stuff` only, the
# function returns the `--verbose` and `--stuff` option items, the
# `do-something` non-option item, three ingested original arguments, and
# three remaining original arguments. This means you can start the next
# argument parsing stage, with option descriptors depending on the
# command name, with the remaining original arguments.
#
# Note that `len(ingested_orig_args)` is NOT always equal to the number
# of returned items, as
#
#     --hello -fdw
#
# for example contains two ingested original arguments, but four
# resulting option items.
#
# On failure, this function raises an `_Error` object.
def parse(orig_args: OrigArgs, opt_descrs: Iterable[OptDescr],
          fail_on_unknown_opt: bool = True) -> _ParseRes:
    # Finds and returns an option description amongst `opt_descrs`
    # having the short option name `short_name` OR the long option name
    # `long_name` (not both).
    def find_opt_descr(short_name: _OptStr = None,
                       long_name: _OptStr = None) -> Optional[OptDescr]:
        for opt_descr in opt_descrs:
            if short_name is not None and short_name == opt_descr.short_name:
                return opt_descr

            if long_name is not None and long_name == opt_descr.long_name:
                return opt_descr

        return None

    # Parses a short option original argument, returning an
    # `_OptParseRes` object.
    #
    # `orig_arg` can contain more than one short options, for example:
    #
    #     -xzv
    #
    # Moreover, `orig_arg` can contain the argument of a short option,
    # for example:
    #
    #     -xzvflol.mp3
    #
    # (`lol.mp3` is the argument of short option `-f`).
    #
    # If this function expects an argument for the last short option of
    # `orig_arg`, then it must be `next_orig_arg`, for example:
    #
    #     -xzvf lol.mp3
    #
    # If any of the short options of `orig_arg` is unknown, then this
    # function raises an error if `fail_on_unknown_opt` is `True`, or
    # returns `None` otherwise.
    def parse_short_opts() -> Optional[_OptParseRes]:
        short_opts = orig_arg[1:]
        items: List[_Item] = []
        done = False
        index = 0
        orig_arg_index_incr = 1

        while not done:
            short_opt = short_opts[index]
            opt_descr = find_opt_descr(short_name=short_opt)

            if opt_descr is None:
                # unknown option
                if fail_on_unknown_opt:
                    raise _Error(orig_arg_index, orig_arg, f'Unknown short option `-{short_opt}`')

                # discard collected arguments
                return None

            opt_arg = None

            if opt_descr.has_arg:
                if index == len(short_opts) - 1:
                    # last short option: use the next original argument
                    if next_orig_arg is None:
                        raise _Error(orig_arg_index, orig_arg,
                                     f'Expecting an argument for short option `-{short_opt}`')

                    opt_arg = next_orig_arg
                    orig_arg_index_incr += 1
                else:
                    # use remaining original argument's text
                    opt_arg = short_opts[index + 1:]

                done = True

            items.append(_OptItem(opt_descr, opt_arg))
            index += 1

            if index == len(short_opts):
                done = True

        return _OptParseRes(items, orig_arg_index_incr)

    # Parses a long option original argument, returning an
    # `_OptParseRes` object.
    #
    # `orig_arg` can contain a single long option, for example:
    #
    #     --header-dir
    #
    # Moreover, `orig_arg` can contain the long option's argument, for
    # example:
    #
    #     --header-dir=/path/to/dir
    #
    # If this function expects an argument for the long option, then it
    # must be `next_orig_arg`, for example:
    #
    #     --header-dir /path/to/dir
    #
    # If the long option is unknown, then this function raises an error
    # if `fail_on_unknown_opt` is `True`, or returns `None` otherwise.
    def parse_long_opt() -> Optional[_OptParseRes]:
        long_opt = orig_arg[2:]
        m = re.match(r'--([^=]+)=(.*)', orig_arg)

        if m:
            # `--long-opt=arg` form: isolate option name
            long_opt = m.group(1)

        opt_descr = find_opt_descr(long_name=long_opt)

        if opt_descr is None:
            # unknown option
            if fail_on_unknown_opt:
                raise _Error(orig_arg_index, orig_arg, f'Unknown long option `--{long_opt}`')

            # discard
            return None

        orig_arg_index_incr = 1

        if opt_descr.has_arg:
            if m:
                item = _OptItem(opt_descr, m.group(2))
            else:
                if next_orig_arg is None:
                    raise _Error(orig_arg_index, orig_arg,
                                 f'Expecting an argument for long option `--{long_opt}`')

                item = _OptItem(opt_descr, next_orig_arg)
                orig_arg_index_incr += 1
        else:
            # no option argument
            item = _OptItem(opt_descr, None)

        return _OptParseRes([item], orig_arg_index_incr)

    # parse original arguments
    items: List[_Item] = []
    orig_arg_index = Index(0)
    non_opt_index = Index(0)

    while orig_arg_index < len(orig_args):
        orig_arg = orig_args[orig_arg_index]

        # keep next original argument, if any
        next_orig_arg = None

        if orig_arg_index < len(orig_args) - 1:
            next_orig_arg = orig_args[orig_arg_index + 1]

        if orig_arg.startswith('-') and len(orig_arg) >= 2:
            # option
            if orig_arg[1] == '-':
                if orig_arg == '--':
                    raise _Error(orig_arg_index, orig_arg, 'Invalid `--` argument')

                # long option
                res = parse_long_opt()
            else:
                # short option(s)
                res = parse_short_opts()

            if res is None:
                # unknown option
                assert not fail_on_unknown_opt
                return _ParseRes(items, orig_args[:orig_arg_index], orig_args[orig_arg_index:])

            items += res.items
            orig_arg_index = Index(orig_arg_index + res.orig_arg_index_incr)
        else:
            # non-option
            items.append(_NonOptItem(orig_arg, orig_arg_index, non_opt_index))
            non_opt_index = Index(non_opt_index + 1)
            orig_arg_index = Index(orig_arg_index + 1)

    return _ParseRes(items, orig_args, [])



# The MIT License (MIT)
#
# Copyright (c) 2015-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import barectf.config_parse_common as barectf_config_parse_common
from barectf.config_parse_common import _ConfigurationParseError
from barectf.config_parse_common import _MapNode
import barectf.config_parse_v2 as barectf_config_parse_v2
import barectf.config_parse_v3 as barectf_config_parse_v3
import barectf.config as barectf_config
import collections
from barectf.typing import Count, VersionNumber
from typing import Optional, List, TextIO
import typing


# Creates and returns a barectf 3 YAML configuration file parser to
# parse the file-like object `file`.
#
# `file` can be a barectf 2 or 3 configuration file.
def _create_v3_parser(file: TextIO, with_pkg_include_dir: bool, include_dirs: Optional[List[str]],
                      ignore_include_not_found: bool) -> barectf_config_parse_v3._Parser:
    try:
        root_node = barectf_config_parse_common._yaml_load(file)

        if type(root_node) is barectf_config_parse_common._ConfigNodeV3:
            # barectf 3 configuration file
            return barectf_config_parse_v3._Parser(file,
                                                   typing.cast(barectf_config_parse_common._ConfigNodeV3,
                                                               root_node),
                                                   with_pkg_include_dir, include_dirs,
                                                   ignore_include_not_found)
        elif type(root_node) is collections.OrderedDict:
            # barectf 2 configuration file
            v2_parser = barectf_config_parse_v2._Parser(file, typing.cast(_MapNode, root_node),
                                                        with_pkg_include_dir, include_dirs,
                                                        ignore_include_not_found)
            return barectf_config_parse_v3._Parser(file, v2_parser.config_node,
                                                   with_pkg_include_dir, include_dirs,
                                                   ignore_include_not_found)
        else:
            raise _ConfigurationParseError('Configuration',
                                           f'Root (configuration) node is not an object (it\'s a `{type(root_node)}`)')
    except _ConfigurationParseError as exc:
        barectf_config_parse_common._append_error_ctx(exc, 'Configuration',
                                                      'Cannot create configuration from YAML file')

        # satisfy static type checker (never reached)
        raise


def _from_file(file: TextIO, with_pkg_include_dir: bool, include_dirs: Optional[List[str]],
               ignore_include_not_found: bool) -> barectf_config.Configuration:
    return _create_v3_parser(file, with_pkg_include_dir, include_dirs, ignore_include_not_found).config


def _effective_config_file(file: TextIO, with_pkg_include_dir: bool,
                           include_dirs: Optional[List[str]], ignore_include_not_found: bool,
                           indent_space_count: Count) -> str:
    config_node = _create_v3_parser(file, with_pkg_include_dir, include_dirs,
                                    ignore_include_not_found).root_node
    return barectf_config_parse_common._yaml_dump(config_node, indent=indent_space_count,
                                                  default_flow_style=False, explicit_start=True,
                                                  explicit_end=True)


def _config_file_major_version(file: TextIO) -> VersionNumber:
    try:
        root_node = barectf_config_parse_common._yaml_load(file)

        if type(root_node) is barectf_config_parse_common._ConfigNodeV3:
            # barectf 3 configuration file
            return VersionNumber(3)
        else:
            # barectf 2 configuration file
            assert type(root_node) is collections.OrderedDict
            return VersionNumber(2)
    except _ConfigurationParseError as exc:
        barectf_config_parse_common._append_error_ctx(exc, 'Configuration', 'Cannot load YAML file')

        # satisfy static type checker (never reached)
        raise



# The MIT License (MIT)
#
# Copyright (c) 2014-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import barectf.template as barectf_template
import barectf.config as barectf_config
import collections
import copy
from typing import List, Optional, Mapping, Callable, Any, Set, Tuple
import typing
from barectf.typing import Count, Alignment


# A tuple containing serialization and size computation function
# templates for a given operation.
_OpTemplates = collections.namedtuple('_OpTemplates', ['serialize', 'size'])


# Abstract base class of any operation within source code.
#
# Any operation has:
#
# * A field type.
#
# * A list of names which, when joined with `_`, form the generic
#   C source variable name.
#
# * A level: how deep this operation is within the operation tree.
#
# * Serialization and size computation templates to generate the
#   operation's source code for those functions.
class _Op:
    def __init__(self, ft: barectf_config._FieldType, names: List[str], level: Count,
                 templates: _OpTemplates):
        self._ft = ft
        self._names = copy.copy(names)
        self._level = level
        self._templates = templates

    @property
    def ft(self) -> barectf_config._FieldType:
        return self._ft

    @property
    def names(self) -> List[str]:
        return self._names

    @property
    def level(self) -> Count:
        return self._level

    @property
    def top_name(self) -> str:
        return self._names[-1]

    def _render_template(self, templ: barectf_template._Template, **kwargs) -> str:
        return templ.render(op=self, root_ft_prefixes=_RootFtPrefixes,
                            root_ft_prefix_names=_ROOT_FT_PREFIX_NAMES, **kwargs)

    def serialize_str(self, **kwargs) -> str:
        return self._render_template(self._templates.serialize, **kwargs)

    def size_str(self, **kwargs) -> str:
        return self._render_template(self._templates.size, **kwargs)


# Compound operation.
#
# A compound operation contains a list of suboperations (leaf or
# compound).
#
# Get the suboperations of a compound operation with its `subops`
# property.
#
# The templates of a compound operation handles its suboperations.
class _CompoundOp(_Op):
    def __init__(self, ft: barectf_config._FieldType, names: List[str], level: Count,
                 templates: _OpTemplates, subops: List[Any] = None):
        super().__init__(ft, names, level, templates)
        self._subops = subops

    @property
    def subops(self):
        return self._subops


# Leaf operation (abstract class).
class _LeafOp(_Op):
    pass


# An "align" operation.
class _AlignOp(_LeafOp):
    def __init__(self, ft: barectf_config._FieldType, names: List[str], level: Count,
                 templates: _OpTemplates, value: Alignment):
        super().__init__(ft, names, level, templates)
        self._value = value

    @property
    def value(self) -> Alignment:
        return self._value


# A "write" operation.
class _WriteOp(_LeafOp):
    def __init__(self, ft: barectf_config._FieldType, names: List[str], level: Count,
                 templates: _OpTemplates, offset_in_byte: Optional[Count]):
        super().__init__(ft, names, level, templates)
        assert offset_in_byte is None or (offset_in_byte >= 0 and offset_in_byte < 8)
        self._offset_in_byte = offset_in_byte

    @property
    def offset_in_byte(self) -> Optional[Count]:
        return self._offset_in_byte


_SpecSerializeWriteTemplates = Mapping[str, barectf_template._Template]


# An operation builder.
#
# Such a builder is closely connected to a `_CodeGen` object using it to
# find generic templates.
#
# Call build_for_root_ft() to make an operation builder create a
# compound operation for a given root structure field type, recursively,
# and return it.
class _OpBuilder:
    def __init__(self, cg: '_CodeGen'):
        self._names: List[str] = []
        self._level = Count(0)
        self._offset_in_byte: Optional[Count] = None
        self._cg = cg

    # Whether or not we're within an array operation.
    @property
    def _in_array(self):
        return self._level > 0

    # Creates and returns an operation for the root structure field type
    # `ft` named `name`.
    #
    # `spec_serialize_write_templates` is a mapping of first level
    # member names to specialized serialization "write" templates.
    def build_for_root_ft(self, ft: barectf_config.StructureFieldType, name: str,
                          spec_serialize_write_templates: Optional[_SpecSerializeWriteTemplates] = None) -> _CompoundOp:
        assert ft is not None

        if spec_serialize_write_templates is None:
            spec_serialize_write_templates = {}

        assert type(ft) is barectf_config.StructureFieldType
        assert len(self._names) == 0
        assert self._level == 0
        ops = self._build_for_ft(ft, name, spec_serialize_write_templates)
        assert len(ops) == 1
        assert type(ops[0]) is _CompoundOp
        return typing.cast(_CompoundOp, ops[0])

    # Creates and returns the operation(s) for a given field type `ft`
    # named `name`.
    #
    # See build_for_root_ft() for `spec_serialize_write_templates`.
    def _build_for_ft(self, ft: barectf_config._FieldType, name: str,
                      spec_serialize_write_templates: _SpecSerializeWriteTemplates) -> List[_Op]:
        def top_name() -> str:
            return self._names[-1]

        # Creates and returns a "write" operation for the field type
        # `ft`.
        #
        # This function considers `spec_serialize_write_templates` to
        # override generic templates.
        def create_write_op(ft: barectf_config._FieldType) -> _WriteOp:
            assert type(ft) is not barectf_config.StructureFieldType
            offset_in_byte = self._offset_in_byte

            if isinstance(ft, barectf_config._BitArrayFieldType) and self._offset_in_byte is not None:
                self._offset_in_byte = Count((self._offset_in_byte + ft.size) % 8)

            serialize_write_templ: Optional[barectf_template._Template] = None

            if len(self._names) == 2:
                serialize_write_templ = spec_serialize_write_templates.get(top_name())

            if serialize_write_templ is None:
                if isinstance(ft, barectf_config._IntegerFieldType):
                    serialize_write_templ = self._cg._serialize_write_int_statements_templ
                elif type(ft) is barectf_config.RealFieldType:
                    serialize_write_templ = self._cg._serialize_write_real_statements_templ
                else:
                    assert type(ft) is barectf_config.StringFieldType
                    serialize_write_templ = self._cg._serialize_write_string_statements_templ

            size_write_templ = None

            if isinstance(ft, barectf_config._BitArrayFieldType):
                size_write_templ = self._cg._size_write_bit_array_statements_templ
            elif type(ft) is barectf_config.StringFieldType:
                size_write_templ = self._cg._size_write_string_statements_templ

            return _WriteOp(ft, self._names, self._level,
                            _OpTemplates(serialize_write_templ, size_write_templ), offset_in_byte)

        # Creates and returns an "align" operation for the field type
        # `ft` if needed.
        #
        # This function updates the builder's state.
        def try_create_align_op(alignment: Alignment, ft: barectf_config._FieldType) -> Optional[_AlignOp]:
            def align(v: Count, alignment: Alignment) -> Count:
                return Count((v + (alignment - 1)) & -alignment)

            if self._offset_in_byte is None and alignment % 8 == 0:
                self._offset_in_byte = Count(0)
            else:
                if self._in_array:
                    self._offset_in_byte = None
                elif self._offset_in_byte is not None:
                    self._offset_in_byte = Count(align(self._offset_in_byte, alignment) % 8)

            if alignment > 1:
                return _AlignOp(ft, self._names, self._level,
                                _OpTemplates(self._cg._serialize_align_statements_templ,
                                             self._cg._size_align_statements_templ),
                                alignment)

            return None

        # Returns whether or not `ft` is a compound field type.
        def ft_is_compound(ft: barectf_config._FieldType) -> bool:
            return isinstance(ft, (barectf_config.StructureFieldType, barectf_config.StaticArrayFieldType))

        # push field type's name to the builder's name stack initially
        self._names.append(name)

        # operations to return
        ops: List[_Op] = []

        if type(ft) is barectf_config.StringFieldType or self._names == [_RootFtPrefixes.PH, 'uuid']:
            # strings and UUID array are always byte-aligned
            op = try_create_align_op(Alignment(8), ft)

            if op is not None:
                ops.append(op)

            ops.append(create_write_op(ft))
        else:
            if ft_is_compound(ft):
                self._offset_in_byte = None

            init_align_op = try_create_align_op(ft.alignment, ft)
            subops: List[_Op] = []

            if type(ft) is barectf_config.StructureFieldType:
                ft = typing.cast(barectf_config.StructureFieldType, ft)

                if init_align_op is not None:
                    # Append structure field's alignment as a
                    # suboperation.
                    #
                    # This is not strictly needed (could be appended to
                    # `ops`), but the properties of `_DsOps` and
                    # `_ErOps` offer a single (structure field type)
                    # operation.
                    subops.append(init_align_op)

                # append suboperations for each member
                for member_name, member in ft.members.items():
                    subops += self._build_for_ft(member.field_type, member_name,
                                                 spec_serialize_write_templates)

                # create structure field's compound operation
                ops.append(_CompoundOp(ft, self._names, self._level,
                                       _OpTemplates(self._cg._serialize_write_struct_statements_templ,
                                                    self._cg._size_write_struct_statements_templ),
                                       subops))
            elif isinstance(ft, barectf_config._ArrayFieldType):
                ft = typing.cast(barectf_config._ArrayFieldType, ft)
                assert ft.alignment == 1 or init_align_op is not None

                if init_align_op is not None:
                    ops.append(init_align_op)

                # append element's suboperations
                self._level = Count(self._level + 1)
                subops += self._build_for_ft(ft.element_field_type,
                                             f'[{_loop_var_name(Count(self._level - 1))}]',
                                             spec_serialize_write_templates)
                self._level = Count(self._level - 1)

                # select the right templates
                if type(ft) is barectf_config.StaticArrayFieldType:
                    templates = _OpTemplates(self._cg._serialize_write_static_array_statements_templ,
                                             self._cg._size_write_static_array_statements_templ)
                else:
                    assert type(ft) is barectf_config.DynamicArrayFieldType
                    templates = _OpTemplates(self._cg._serialize_write_dynamic_array_statements_templ,
                                             self._cg._size_write_dynamic_array_statements_templ)

                # create array field's compound operation
                ops.append(_CompoundOp(ft, self._names, self._level, templates, subops))
            else:
                # leaf field: align + write
                if init_align_op is not None:
                    ops.append(init_align_op)

                ops.append(create_write_op(ft))

        # exiting for this field type: pop its name
        del self._names[-1]

        return ops


_OptCompoundOp = Optional[_CompoundOp]


# The operations for an event record.
#
# The available operations are:
#
# * Specific context operation.
# * Payload operation.
class _ErOps:
    def __init__(self, spec_ctx_op: _OptCompoundOp, payload_op: _OptCompoundOp):
        self._spec_ctx_op = spec_ctx_op
        self._payload_op = payload_op

    @property
    def spec_ctx_op(self) -> _OptCompoundOp:
        return self._spec_ctx_op

    @property
    def payload_op(self) -> _OptCompoundOp:
        return self._payload_op


_ErOpsMap = Mapping[barectf_config.EventRecordType, _ErOps]


# The operations for a data stream.
#
# The available operations are:
#
# * Packet header operation.
# * Packet context operation.
# * Event record header operation.
# * Event record common context operation.
# * Event record operations (`_ErOps`).
class _DsOps:
    def __init__(self, pkt_header_op: _OptCompoundOp, pkt_ctx_op: _CompoundOp,
                 er_header_op: _OptCompoundOp, er_common_ctx_op: _OptCompoundOp, er_ops: _ErOpsMap):
        self._pkt_header_op = pkt_header_op
        self._pkt_ctx_op = pkt_ctx_op
        self._er_header_op = er_header_op
        self._er_common_ctx_op = er_common_ctx_op
        self._er_ops = er_ops

    @property
    def pkt_header_op(self) -> _OptCompoundOp:
        return self._pkt_header_op

    @property
    def pkt_ctx_op(self) -> _CompoundOp:
        return self._pkt_ctx_op

    @property
    def er_header_op(self) -> _OptCompoundOp:
        return self._er_header_op

    @property
    def er_common_ctx_op(self) -> _OptCompoundOp:
        return self._er_common_ctx_op

    @property
    def er_ops(self) -> _ErOpsMap:
        return self._er_ops


# The C variable name prefixes for the six kinds of root field types.
class _RootFtPrefixes:
    PH = 'ph'
    PC = 'pc'
    ERH = 'h'
    ERCC = 'cc'
    ERSC = 'sc'
    ERP = 'p'


# The human-readable names of the `_RootFtPrefixes` members.
_ROOT_FT_PREFIX_NAMES = {
    _RootFtPrefixes.PH: 'packet header',
    _RootFtPrefixes.PC: 'packet context',
    _RootFtPrefixes.ERH: 'header',
    _RootFtPrefixes.ERCC: 'common context',
    _RootFtPrefixes.ERSC: 'specific context',
    _RootFtPrefixes.ERP: 'payload',
}


# A named function parameter for a given field type.
_FtParam = collections.namedtuple('_FtParam', ['ft', 'name'])


# C type abstract base class.
class _CType:
    def __init__(self, is_const: bool):
        self._is_const = is_const

    @property
    def is_const(self) -> bool:
        return self._is_const


# Arithmetic C type.
class _ArithCType(_CType):
    def __init__(self, name: str, is_const: bool):
        super().__init__(is_const)
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def __str__(self) -> str:
        return f'{"const " if self._is_const else ""}{self._name}'


# Pointer C type.
class _PointerCType(_CType):
    def __init__(self, pointed_c_type: _CType, is_const: bool):
        super().__init__(is_const)
        self._pointed_c_type = pointed_c_type

    @property
    def pointed_c_type(self) -> _CType:
        return self._pointed_c_type

    def __str__(self) -> str:
        s = str(self._pointed_c_type)

        if not s.endswith('*'):
            s += ' '

        s += '*'

        if self._is_const:
            s += ' const'

        return s


# Returns the name of a loop variable given a nesting level `level`.
def _loop_var_name(level: Count) -> str:
    if level < 3:
        return 'ijk'[level]

    return f'k{level - 2}'


# A C code generator.
#
# Such a code generator can generate:
#
# * The bitfield header (gen_bitfield_header()).
# * The public header (gen_header()).
# * The source code (gen_src()).
class _CodeGen:
    def __init__(self, cfg: barectf_config.Configuration):
        self._cfg = cfg
        self._iden_prefix = cfg.options.code_generation_options.identifier_prefix
        self._templ_filters: Mapping[str, Callable[..., Any]] = {
            'ft_c_type': self._ft_c_type,
            'open_func_params_str': self._open_func_params_str,
            'trace_func_params_str': self._trace_func_params_str,
            'serialize_er_common_ctx_func_params_str': self._serialize_er_common_ctx_func_params_str,
            'loop_var_name': _loop_var_name,
            'op_src_var_name': self._op_src_var_name,
        }
        self._func_proto_params_templ = self._create_template('func-proto-params.j2')
        self._serialize_align_statements_templ = self._create_template('serialize-align-statements.j2')
        self._serialize_write_int_statements_templ = self._create_template('serialize-write-int-statements.j2')
        self._serialize_write_real_statements_templ = self._create_template('serialize-write-real-statements.j2')
        self._serialize_write_string_statements_templ = self._create_template('serialize-write-string-statements.j2')
        self._serialize_write_struct_statements_templ = self._create_template('serialize-write-struct-statements.j2')
        self._serialize_write_static_array_statements_templ = self._create_template('serialize-write-static-array-statements.j2')
        self._serialize_write_dynamic_array_statements_templ = self._create_template('serialize-write-dynamic-array-statements.j2')
        self._serialize_write_magic_statements_templ = self._create_template('serialize-write-magic-statements.j2')
        self._serialize_write_uuid_statements_templ = self._create_template('serialize-write-uuid-statements.j2')
        self._serialize_write_dst_id_statements_templ = self._create_template('serialize-write-dst-id-statements.j2')
        self._serialize_write_timestamp_statements_templ = self._create_template('serialize-write-timestamp-statements.j2')
        self._serialize_write_packet_size_statements_templ = self._create_template('serialize-write-packet-size-statements.j2')
        self._serialize_write_seq_num_statements_templ = self._create_template('serialize-write-seq-num-statements.j2')
        self._serialize_write_skip_save_statements_templ = self._create_template('serialize-write-skip-save-statements.j2')
        self._serialize_write_ert_id_statements_templ = self._create_template('serialize-write-ert-id-statements.j2')
        self._size_align_statements_templ = self._create_template('size-align-statements.j2')
        self._size_write_bit_array_statements_templ = self._create_template('size-write-bit-array-statements.j2')
        self._size_write_string_statements_templ = self._create_template('size-write-string-statements.j2')
        self._size_write_struct_statements_templ = self._create_template('size-write-struct-statements.j2')
        self._size_write_static_array_statements_templ = self._create_template('size-write-static-array-statements.j2')
        self._size_write_dynamic_array_statements_templ = self._create_template('size-write-dynamic-array-statements.j2')

    # Creates and returns a template named `name` which is a file
    # template if `is_file_template` is `True`.
    #
    # `name` is the file name, including the `.j2` extension, within the
    # `c` directory.
    #
    # Such a template has the filters custom filters
    # `self._templ_filters`.
    def _create_template_base(self, name: str,
                              is_file_template: bool) -> barectf_template._Template:
        return barectf_template._Template(f'c/{name}', is_file_template, self._cfg,
                                          self._templ_filters)

    # Creates and returns a non-file template named `name`.
    #
    # See _create_template_base() for `name`.
    def _create_template(self, name: str) -> barectf_template._Template:
        return self._create_template_base(name, False)

    # Creates and returns a file template named `name`.
    #
    # See _create_template_base() for `name`.
    def _create_file_template(self, name: str) -> barectf_template._Template:
        return self._create_template_base(name, True)

    # Trace type of this code generator's barectf configuration.
    @property
    def _trace_type(self) -> barectf_config._TraceType:
        return self._cfg.trace.type

    # Returns the name of a source variable for the operation `op`.
    def _op_src_var_name(self, op: _LeafOp) -> str:
        s = ''

        for index, name in enumerate(op.names):
            if index > 0 and not name.startswith('['):
                s += '_'

            s += name

        return s

    # Returns the C type for the field type `ft`, making it `const` if
    # `is_const` is `True`.
    def _ft_c_type(self, ft: barectf_config._FieldType, is_const: bool = False):
        if isinstance(ft, barectf_config._IntegerFieldType):
            ft = typing.cast(barectf_config._IntegerFieldType, ft)
            sign_prefix = 'u' if isinstance(ft, barectf_config.UnsignedIntegerFieldType) else ''

            if ft.size <= 8:
                sz = 8
            elif ft.size <= 16:
                sz = 16
            elif ft.size <= 32:
                sz = 32
            else:
                assert ft.size <= 64
                sz = 64

            return _ArithCType(f'{sign_prefix}int{sz}_t', is_const)
        elif type(ft) is barectf_config.RealFieldType:
            ft = typing.cast(barectf_config.RealFieldType, ft)

            if ft.size == 32 and ft.alignment == 32:
                s = 'float'
            elif ft.size == 64 and ft.alignment == 64:
                s = 'double'
            else:
                s = 'uint64_t'

            return _ArithCType(s, is_const)
        elif type(ft) is barectf_config.StringFieldType:
            return _PointerCType(_ArithCType('char', True), is_const)
        else:
            assert isinstance(ft, barectf_config._ArrayFieldType)
            ft = typing.cast(barectf_config._ArrayFieldType, ft)
            return _PointerCType(self._ft_c_type(ft.element_field_type, True), is_const)

    # Returns the function prototype parameters for the members of the
    # root structure field type `root_ft`.
    #
    # Each parameter has the prefix `name_prefix` followed with `_`.
    #
    # Members of which the name is in `exclude_set` are excluded.
    def _proto_params_str(self, root_ft: Optional[barectf_config.StructureFieldType],
                          name_prefix: str, const_params: bool,
                          exclude_set: Optional[Set[str]] = None, only_dyn: bool = False) -> str:
        if root_ft is None:
            return ''

        if exclude_set is None:
            exclude_set = set()

        params = []

        for member_name, member in root_ft.members.items():
            if member_name in exclude_set:
                continue

            is_dyn = member.field_type.size_is_dynamic

            if isinstance(member.field_type, barectf_config.UnsignedIntegerFieldType):
                ft = typing.cast(barectf_config.UnsignedIntegerFieldType, member.field_type)
                is_dyn = is_dyn or ft._is_len

            if only_dyn and not is_dyn:
                continue

            params.append(_FtParam(member.field_type, member_name))

        return self._func_proto_params_templ.render(params=params, prefix=name_prefix,
                                                    const_params=const_params)

    # Returns the packet opening function prototype parameters for the
    # data stream type `dst`.
    def _open_func_params_str(self, dst: barectf_config.DataStreamType, const_params: bool) -> str:
        parts = []
        parts.append(self._proto_params_str(self._trace_type._pkt_header_ft, _RootFtPrefixes.PH,
                                            const_params, {'magic', 'stream_id', 'uuid'}))

        exclude_set = {
            'timestamp_begin',
            'timestamp_end',
            'packet_size',
            'content_size',
            'events_discarded',
            'packet_seq_num',
        }
        parts.append(self._proto_params_str(dst._pkt_ctx_ft, _RootFtPrefixes.PC, const_params,
                                            exclude_set))
        return ''.join(parts)

    # Returns the tracing function prototype parameters for the data
    # stream and event record types `ds_er_types`.
    def _trace_func_params_str(self, ds_er_types: Tuple[barectf_config.DataStreamType,
                                                        barectf_config.EventRecordType],
                               const_params: bool, only_dyn: bool = False):
        dst = ds_er_types[0]
        ert = ds_er_types[1]
        parts = []

        if dst._er_header_ft is not None:
            parts.append(self._proto_params_str(dst._er_header_ft, _RootFtPrefixes.ERH,
                                                const_params, {'id', 'timestamp'},
                                                only_dyn=only_dyn))

        if dst.event_record_common_context_field_type is not None:
            parts.append(self._proto_params_str(dst.event_record_common_context_field_type,
                                                _RootFtPrefixes.ERCC, const_params,
                                                only_dyn=only_dyn))

        if ert.specific_context_field_type is not None:
            parts.append(self._proto_params_str(ert.specific_context_field_type,
                                                _RootFtPrefixes.ERSC, const_params,
                                                only_dyn=only_dyn))

        if ert.payload_field_type is not None:
            parts.append(self._proto_params_str(ert.payload_field_type, _RootFtPrefixes.ERP,
                                                const_params, only_dyn=only_dyn))

        return ''.join(parts)

    # Returns the event record common context serialization function
    # prototype parameters for the data stream type `dst`.
    def _serialize_er_common_ctx_func_params_str(self, dst: barectf_config.DataStreamType,
                                                 const_params: bool) -> str:
        return self._proto_params_str(dst.event_record_common_context_field_type,
                                      _RootFtPrefixes.ERCC, const_params)

    # Generates the bitfield header file contents.
    def gen_bitfield_header(self) -> str:
        return self._create_file_template('bitfield.h.j2').render()

    # Generates the public header file contents.
    def gen_header(self) -> str:
        return self._create_file_template('barectf.h.j2').render(root_ft_prefixes=_RootFtPrefixes)

    # Generates the source code file contents.
    def gen_src(self, header_file_name: str, bitfield_header_file_name: str) -> str:
        # Creates and returns the operations for all the data stream and
        # for all their event records.
        def create_ds_ops() -> Mapping[barectf_config.DataStreamType, _DsOps]:
            ds_ops = {}

            for dst in self._trace_type.data_stream_types:
                pkt_header_op = None
                builder = _OpBuilder(self)
                pkt_header_ft = self._trace_type._pkt_header_ft

                # packet header operations
                if pkt_header_ft is not None:
                    spec_serialize_write_templates = {
                        'magic': self._serialize_write_magic_statements_templ,
                        'uuid': self._serialize_write_uuid_statements_templ,
                        'stream_id': self._serialize_write_dst_id_statements_templ,
                    }
                    pkt_header_op = builder.build_for_root_ft(pkt_header_ft,
                                                                  _RootFtPrefixes.PH,
                                                                  spec_serialize_write_templates)

                # packet context operation
                spec_serialize_write_templates = {
                    'timestamp_begin': self._serialize_write_timestamp_statements_templ,
                    'packet_size': self._serialize_write_packet_size_statements_templ,
                    'timestamp_end': self._serialize_write_skip_save_statements_templ,
                    'events_discarded': self._serialize_write_skip_save_statements_templ,
                    'content_size': self._serialize_write_skip_save_statements_templ,
                    'packet_seq_num': self._serialize_write_seq_num_statements_templ,
                }
                pkt_ctx_op = builder.build_for_root_ft(dst._pkt_ctx_ft, _RootFtPrefixes.PC,
                                                       spec_serialize_write_templates)

                # event record header operation
                builder = _OpBuilder(self)
                er_header_op = None

                if dst._er_header_ft is not None:
                    spec_serialize_write_templates = {
                        'timestamp': self._serialize_write_timestamp_statements_templ,
                        'id': self._serialize_write_ert_id_statements_templ,
                    }
                    er_header_op = builder.build_for_root_ft(dst._er_header_ft, _RootFtPrefixes.ERH,
                                                             spec_serialize_write_templates)

                # event record common context operation
                er_common_ctx_op = None

                if dst.event_record_common_context_field_type is not None:
                    er_common_ctx_op = builder.build_for_root_ft(dst.event_record_common_context_field_type,
                                                                 _RootFtPrefixes.ERCC)

                # operations specific to each event record type
                er_ops = {}

                for ert in dst.event_record_types:
                    ev_builder = copy.copy(builder)

                    # specific context operation
                    spec_ctx_op = None

                    if ert.specific_context_field_type is not None:
                        spec_ctx_op = ev_builder.build_for_root_ft(ert.specific_context_field_type,
                                                                   _RootFtPrefixes.ERSC)

                    # payload operation
                    payload_op = None

                    if ert.payload_field_type is not None:
                        payload_op = ev_builder.build_for_root_ft(ert.payload_field_type,
                                                                  _RootFtPrefixes.ERP)

                    er_ops[ert] = _ErOps(spec_ctx_op, payload_op)

                ds_ops[dst] = _DsOps(pkt_header_op, pkt_ctx_op, er_header_op, er_common_ctx_op,
                                     er_ops)

            return ds_ops

        # Returns the "write" operation for the packet context member
        # named `member_name` within the data stream type `dst`.
        def ds_op_pkt_ctx_op(dst: barectf_config.DataStreamType, member_name: str) -> _Op:
            ret_op = None

            for op in ds_ops[dst].pkt_ctx_op.subops:
                if op.top_name == member_name and type(op) is _WriteOp:
                    ret_op = op
                    break

            assert ret_op is not None
            return typing.cast(_Op, ret_op)

        ds_ops = create_ds_ops()
        return self._create_file_template('barectf.c.j2').render(header_file_name=header_file_name,
                                                                 bitfield_header_file_name=bitfield_header_file_name,
                                                                 root_ft_prefixes=_RootFtPrefixes,
                                                                 root_ft_prefix_names=_ROOT_FT_PREFIX_NAMES,
                                                                 ds_ops=ds_ops,
                                                                 ds_op_pkt_ctx_op=ds_op_pkt_ctx_op)



# The MIT License (MIT)
#
# Copyright (c) 2014-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import barectf.tsdl182gen as barectf_tsdl182gen
import barectf.config as barectf_config
import barectf.cgen as barectf_cgen
from typing import List, Optional


# A file generated by a `CodeGenerator` object.
#
# A generated file has a name (influenced by the configuration's
# file name prefix option) and contents.
class _GeneratedFile:
    def __init__(self, name: str, contents: str):
        self._name = name
        self._contents = contents

    @property
    def name(self) -> str:
        return self._name

    @property
    def contents(self) -> str:
        return self._contents


# A barectf code generator.
#
# Build a code generator with a barectf configuration.
#
# A code generator can generate the TSDL `metadata` file and C source
# and header files.
class CodeGenerator:
    def __init__(self, configuration: barectf_config.Configuration):
        self._config = configuration
        self._file_name_prefix = configuration.options.code_generation_options.file_name_prefix
        self._c_code_gen = barectf_cgen._CodeGen(configuration)
        self._c_headers: Optional[List[_GeneratedFile]] = None
        self._c_sources: Optional[List[_GeneratedFile]] = None
        self._metadata_stream: Optional[_GeneratedFile] = None

    @property
    def _barectf_header_name(self) -> str:
        return f'{self._file_name_prefix}.h'

    @property
    def _bitfield_header_name(self) -> str:
        return f'{self._file_name_prefix}-bitfield.h'

    def generate_c_headers(self) -> List[_GeneratedFile]:
        if self._c_headers is None:
            self._c_headers = [
                _GeneratedFile(self._barectf_header_name, self._c_code_gen.gen_header()),
                _GeneratedFile(self._bitfield_header_name, self._c_code_gen.gen_bitfield_header()),
            ]

        return self._c_headers

    def generate_c_sources(self) -> List[_GeneratedFile]:
        if self._c_sources is None:
            self._c_sources = [
                _GeneratedFile(f'{self._file_name_prefix}.c',
                               self._c_code_gen.gen_src(self._barectf_header_name,
                                                        self._bitfield_header_name))
            ]

        return self._c_sources

    def generate_metadata_stream(self) -> _GeneratedFile:
        if self._metadata_stream is None:
            self._metadata_stream = _GeneratedFile('metadata',
                                                   barectf_tsdl182gen._from_config(self._config))

        return self._metadata_stream



# The MIT License (MIT)
#
# Copyright (c) 2015-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import barectf.config_parse_common as barectf_config_parse_common
from barectf.config_parse_common import _ConfigurationParseError
from barectf.config_parse_common import _append_error_ctx
from barectf.config_parse_common import _MapNode
import barectf.config as barectf_config
from barectf.config import _OptStructFt
import collections
import uuid
from barectf.typing import Count, Alignment, VersionNumber
from typing import Optional, List, Dict, Any, TextIO, Set, Iterable, Callable, Tuple, Type
import typing


# A barectf 3 YAML configuration parser.
#
# When you build such a parser, it parses the configuration node `node`
# (already loaded from the file having the path `path`) and creates a
# corresponding `barectf.Configuration` object which you can get with
# the `config` property.
#
# See the comments of _parse() for more implementation details about the
# parsing stages and general strategy.
class _Parser(barectf_config_parse_common._Parser):
    # Builds a barectf 3 YAML configuration parser and parses the root
    # configuration node `node` (already loaded from the file-like
    # object `root_file`).
    def __init__(self, root_file: TextIO, node: barectf_config_parse_common._ConfigNodeV3,
                 with_pkg_include_dir: bool, inclusion_dirs: Optional[List[str]],
                 ignore_include_not_found: bool):
        super().__init__(root_file, node, with_pkg_include_dir, inclusion_dirs,
                         ignore_include_not_found, VersionNumber(3))
        self._ft_cls_name_to_create_method: Dict[str, Callable[[_MapNode],
                                                               List[barectf_config._FieldType]]] = {
            'unsigned-integer': self._create_int_ft,
            'signed-integer': self._create_int_ft,
            'unsigned-enumeration': self._create_enum_ft,
            'signed-enumeration': self._create_enum_ft,
            'real': self._create_real_ft,
            'string': self._create_string_ft,
            'static-array': self._create_static_array_ft,
            'dynamic-array': self._create_dynamic_array_ft,
            'structure': self._create_struct_ft,
        }
        self._parse()

    # Validates the alignment `alignment`, raising a
    # `_ConfigurationParseError` exception using `ctx_obj_name` if it's
    # invalid.
    @staticmethod
    def _validate_alignment(alignment: Alignment, ctx_obj_name: str):
        assert alignment >= 1

        # check for power of two
        if (alignment & (alignment - 1)) != 0:
            raise _ConfigurationParseError(ctx_obj_name,
                                           f'Invalid alignment (not a power of two): {alignment}')

    # Validates the TSDL identifier `iden`, raising a
    # `_ConfigurationParseError` exception using `ctx_obj_name` and
    # `prop` to format the message if it's invalid.
    @staticmethod
    def _validate_iden(iden: str, ctx_obj_name: str, prop: str):
        assert type(iden) is str
        ctf_keywords = {
            'align',
            'callsite',
            'clock',
            'enum',
            'env',
            'event',
            'floating_point',
            'integer',
            'stream',
            'string',
            'struct',
            'trace',
            'typealias',
            'typedef',
            'variant',
        }

        if iden in ctf_keywords:
            msg = f'Invalid {prop} (not a valid identifier): `{iden}`'
            raise _ConfigurationParseError(ctx_obj_name, msg)

    @staticmethod
    def _alignment_prop(ft_node: _MapNode, prop_name: str) -> Alignment:
        alignment = ft_node.get(prop_name)

        if alignment is not None:
            _Parser._validate_alignment(alignment, '`prop_name` property')

        return Alignment(alignment)

    @property
    def _trace_type_node(self) -> _MapNode:
        return self.config_node['trace']['type']

    @staticmethod
    def _byte_order_from_node(node: str) -> barectf_config.ByteOrder:
        return {
            'big-endian': barectf_config.ByteOrder.BIG_ENDIAN,
            'little-endian': barectf_config.ByteOrder.LITTLE_ENDIAN,
        }[node]

    # Creates a bit array field type having the type `ft_type` from the
    # bit array field type node `ft_node`, passing the additional
    # `*args` to ft_type.__init__().
    def _create_common_bit_array_ft(self, ft_node: _MapNode,
                                    ft_type: Type[barectf_config._BitArrayFieldType],
                                    default_alignment: Optional[Alignment],
                                    *args) -> barectf_config._BitArrayFieldType:
        alignment = self._alignment_prop(ft_node, 'alignment')

        if alignment is None:
            alignment = default_alignment

        return ft_type(ft_node['size'], alignment, *args)

    # Creates an integer field type having the type `ft_type` from the
    # integer field type node `ft_node`, passing the additional `*args`
    # to ft_type.__init__().
    def _create_common_int_ft(self, ft_node: _MapNode,
                              ft_type: Type[barectf_config._IntegerFieldType], *args) -> barectf_config._IntegerFieldType:
        preferred_display_base = {
            'binary': barectf_config.DisplayBase.BINARY,
            'octal': barectf_config.DisplayBase.OCTAL,
            'decimal': barectf_config.DisplayBase.DECIMAL,
            'hexadecimal': barectf_config.DisplayBase.HEXADECIMAL,
        }[ft_node.get('preferred-display-base', 'decimal')]
        return typing.cast(barectf_config._IntegerFieldType,
                           self._create_common_bit_array_ft(ft_node, ft_type, None,
                                                            preferred_display_base, *args))

    # Creates an integer field type from the unsigned/signed integer
    # field type node `ft_node`.
    def _create_int_ft(self, ft_node: _MapNode) -> List[barectf_config._FieldType]:
        ft_type = {
            'unsigned-integer': barectf_config.UnsignedIntegerFieldType,
            'signed-integer': barectf_config.SignedIntegerFieldType,
        }[ft_node['class']]
        return [self._create_common_int_ft(ft_node, ft_type)]

    # Creates an enumeration field type from the unsigned/signed
    # enumeration field type node `ft_node`.
    def _create_enum_ft(self, ft_node: _MapNode) -> List[barectf_config._FieldType]:
        ft_type = {
            'unsigned-enumeration': barectf_config.UnsignedEnumerationFieldType,
            'signed-enumeration': barectf_config.SignedEnumerationFieldType,
        }[ft_node['class']]
        mappings = collections.OrderedDict()

        for label, mapping_node in ft_node['mappings'].items():
            ranges = set()

            for range_node in mapping_node:
                if type(range_node) is list:
                    ranges.add(barectf_config.EnumerationFieldTypeMappingRange(range_node[0],
                                                                               range_node[1]))
                else:
                    assert type(range_node) is int
                    ranges.add(barectf_config.EnumerationFieldTypeMappingRange(range_node,
                                                                               range_node))

            mappings[label] = barectf_config.EnumerationFieldTypeMapping(ranges)

        return [typing.cast(barectf_config._EnumerationFieldType,
                            self._create_common_int_ft(ft_node, ft_type,
                                                       barectf_config.EnumerationFieldTypeMappings(mappings)))]

    # Creates a real field type from the real field type node `ft_node`.
    def _create_real_ft(self, ft_node: _MapNode) -> List[barectf_config._FieldType]:
        return [typing.cast(barectf_config.RealFieldType,
                            self._create_common_bit_array_ft(ft_node, barectf_config.RealFieldType,
                                                             Alignment(8)))]

    # Creates a string field type from the string field type node
    # `ft_node`.
    def _create_string_ft(self, ft_node: _MapNode) -> List[barectf_config._FieldType]:
        return [barectf_config.StringFieldType()]

    def _create_array_ft(self, ft_type, ft_node: _MapNode, **kwargs) -> barectf_config._ArrayFieldType:
        prop_name = 'element-field-type'

        try:
            element_fts = self._create_fts(ft_node[prop_name])
        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, f'`{prop_name}` property')

        if len(element_fts) != 1 or isinstance(element_fts[0], (barectf_config.StructureFieldType,
                                                                barectf_config.DynamicArrayFieldType)):
            raise _ConfigurationParseError(f'`{prop_name}` property',
                                           'Nested structure and dynamic array field types are not supported')

        return ft_type(element_field_type=element_fts[0], **kwargs)

    # Creates a static array field type from the static array field type
    # node `ft_node`.
    def _create_static_array_ft(self, ft_node: _MapNode) -> List[barectf_config._FieldType]:
        return [typing.cast(barectf_config.StaticArrayFieldType,
                            self._create_array_ft(barectf_config.StaticArrayFieldType, ft_node,
                                                  length=ft_node['length']))]

    # Creates a dynamic array field type from the dynamic array field
    # type node `ft_node`.
    def _create_dynamic_array_ft(self, ft_node: _MapNode) -> List[barectf_config._FieldType]:
        # create length unsigned integer field type
        len_ft = barectf_config.UnsignedIntegerFieldType(32, alignment=Alignment(8))
        return [
            len_ft,
            typing.cast(barectf_config.DynamicArrayFieldType,
                        self._create_array_ft(barectf_config.DynamicArrayFieldType, ft_node,
                                              length_field_type=len_ft))
        ]

    # Creates structure field type members from the structure field type
    # members node `members_node`.
    #
    # `prop_name` is the name of the property of which `members_node` is
    # the value.
    def _create_struct_ft_members(self, members_node: List[_MapNode], prop_name: str):
        members = collections.OrderedDict()
        member_names: Set[str] = set()

        for member_node in members_node:
            member_name, member_node = list(member_node.items())[0]

            if member_name in member_names:
                raise _ConfigurationParseError(f'`{prop_name}` property',
                                               f'Duplicate member `{member_name}`')

            self._validate_iden(member_name, f'`{prop_name}` property',
                                'structure field type member name')
            member_names.add(member_name)
            ft_prop_name = 'field-type'
            ft_node = member_node[ft_prop_name]

            try:
                if ft_node['class'] in ['structure']:
                    raise _ConfigurationParseError(f'`{ft_prop_name}` property',
                                                   'Nested structure field types are not supported')

                try:
                    member_fts = self._create_fts(ft_node)
                except _ConfigurationParseError as exc:
                    _append_error_ctx(exc, f'`{ft_prop_name}` property')
            except _ConfigurationParseError as exc:
                _append_error_ctx(exc, f'Structure field type member `{member_name}`')

            if len(member_fts) == 2:
                # The only case where this happens is a dynamic array
                # field type node which generates an unsigned integer
                # field type for the length and the dynamic array field
                # type itself.
                assert type(member_fts[1]) is barectf_config.DynamicArrayFieldType
                members[f'__{member_name}_len'] = barectf_config.StructureFieldTypeMember(member_fts[0])
            else:
                assert len(member_fts) == 1

            members[member_name] = barectf_config.StructureFieldTypeMember(member_fts[-1])

        return barectf_config.StructureFieldTypeMembers(members)

    # Creates a structure field type from the structure field type node
    # `ft_node`.
    def _create_struct_ft(self, ft_node: _MapNode) -> List[barectf_config._FieldType]:
        minimum_alignment = self._alignment_prop(ft_node, 'minimum-alignment')

        if minimum_alignment is None:
            minimum_alignment = 1

        members = None
        prop_name = 'members'
        members_node = ft_node.get(prop_name)

        if members_node is not None:
            members = self._create_struct_ft_members(members_node, prop_name)

        return [barectf_config.StructureFieldType(minimum_alignment, members)]

    # Creates field types from the field type node `ft_node`.
    def _create_fts(self, ft_node: _MapNode) -> List[barectf_config._FieldType]:
        return self._ft_cls_name_to_create_method[ft_node['class']](ft_node)

    # Creates field types from the field type node `parent_node[key]`
    # if it exists.
    def _try_create_fts(self, parent_node: _MapNode, key: str) -> Optional[List[barectf_config._FieldType]]:
        if key not in parent_node:
            return None

        try:
            return self._create_fts(parent_node[key])
        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, f'`{key}` property')

            # satisfy static type checker (never reached)
            raise

    # Like _try_create_fts(), but casts the result's type (first and
    # only element) to `barectf_config.StructureFieldType` to satisfy
    # static type checkers.
    def _try_create_struct_ft(self, parent_node: _MapNode, key: str) -> _OptStructFt:
        fts = self._try_create_fts(parent_node, key)

        if fts is None:
            return None

        return typing.cast(barectf_config.StructureFieldType, fts[0])

    # Returns the total number of members in the structure field type
    # node `ft_node` if it exists, otherwise 0.
    @staticmethod
    def _total_struct_ft_node_members(ft_node: Optional[_MapNode]) -> Count:
        if ft_node is None:
            return Count(0)

        members_node = ft_node.get('members')

        if members_node is None:
            return Count(0)

        return Count(len(members_node))

    # Creates an event record type from the event record type node
    # `ert_node` named `name`.
    #
    # `ert_member_count` is the total number of structure field type
    # members within the event record type so far (from the common part
    # in its data stream type). For example, if the data stream type has
    # an event record header field type with `id` and `timestamp`
    # members, then `ert_member_count` is 2.
    def _create_ert(self, name: str, ert_node: _MapNode,
                        ert_member_count: Count) -> barectf_config.EventRecordType:
        try:
            self._validate_iden(name, '`name` property', 'event record type name')

            # make sure the event record type is not empty
            spec_ctx_ft_prop_name = 'specific-context-field-type'
            payload_ft_prop_name = 'payload-field-type'
            ert_member_count = Count(ert_member_count + self._total_struct_ft_node_members(ert_node.get(spec_ctx_ft_prop_name)))
            ert_member_count = Count(ert_member_count + self._total_struct_ft_node_members(ert_node.get(payload_ft_prop_name)))

            if ert_member_count == 0:
                raise _ConfigurationParseError('Event record type',
                                               'Event record type is empty (no members).')

            # create event record type
            return barectf_config.EventRecordType(name, ert_node.get('log-level'),
                                                  self._try_create_struct_ft(ert_node,
                                                                       spec_ctx_ft_prop_name),
                                                  self._try_create_struct_ft(ert_node,
                                                                       payload_ft_prop_name))
        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, f'Event record type `{name}`')

            # satisfy static type checker (never reached)
            raise

    # Returns the effective feature field type for the field type
    # node `parent_node[key]`, if any.
    #
    # Returns:
    #
    # If `parent_node[key]` is `False`:
    #     `None`.
    #
    # If `parent_node[key]` is `True`:
    #     `barectf_config.DEFAULT_FIELD_TYPE`.
    #
    # If `parent_node[key]` doesn't exist:
    #     `none` (parameter).
    #
    # Otherwise:
    #     A created field type.
    def _feature_ft(self, parent_node: _MapNode, key: str, none: Any = None) -> Any:
        if key not in parent_node:
            # missing: default feature field type
            return none

        ft_node = parent_node[key]
        assert ft_node is not None

        if ft_node is True:
            # default feature field type
            return barectf_config.DEFAULT_FIELD_TYPE

        if ft_node is False:
            # disabled feature
            return None

        assert type(ft_node) is collections.OrderedDict
        return self._create_fts(ft_node)[0]

    def _create_dst(self, name: str, dst_node: _MapNode) -> barectf_config.DataStreamType:
        try:
            # validate data stream type's name
            self._validate_iden(name, '`name` property', 'data stream type name')

            # get default clock type, if any
            def_clk_type = None
            prop_name = '$default-clock-type-name'
            def_clk_type_name = dst_node.get(prop_name)

            if def_clk_type_name is not None:
                try:
                    def_clk_type = self._clk_type(def_clk_type_name, prop_name)
                except _ConfigurationParseError as exc:
                    _append_error_ctx(exc, f'`{prop_name}` property')

            # create feature field types
            pkt_total_size_ft = barectf_config.DEFAULT_FIELD_TYPE
            pkt_content_size_ft = barectf_config.DEFAULT_FIELD_TYPE
            pkt_beginning_ts_ft = None
            pkt_end_ts_ft = None
            pkt_disc_er_counter_snap_ft = barectf_config.DEFAULT_FIELD_TYPE
            ert_id_ft = barectf_config.DEFAULT_FIELD_TYPE
            ert_ts_ft = None
            pkt_seq_num_ft = None

            if def_clk_type is not None:
                # The data stream type has a default clock type.
                # Initialize the packet beginning timestamp, packet end
                # timestamp, and event record timestamp field types to
                # default field types.
                #
                # This means your data stream type node only needs a
                # default clock type name to enable those features
                # automatically. Those features do not add any parameter
                # to the event tracing functions.
                pkt_beginning_ts_ft = barectf_config.DEFAULT_FIELD_TYPE
                pkt_end_ts_ft = barectf_config.DEFAULT_FIELD_TYPE
                ert_ts_ft = barectf_config.DEFAULT_FIELD_TYPE

            features_node = dst_node.get('$features')

            if features_node is not None:
                # create packet feature field types
                pkt_node = features_node.get('packet')

                if pkt_node is not None:
                    pkt_total_size_ft = self._feature_ft(pkt_node, 'total-size-field-type',
                                                         pkt_total_size_ft)
                    pkt_content_size_ft = self._feature_ft(pkt_node, 'content-size-field-type',
                                                           pkt_content_size_ft)
                    pkt_beginning_ts_ft = self._feature_ft(pkt_node,
                                                           'beginning-timestamp-field-type',
                                                           pkt_beginning_ts_ft)
                    pkt_end_ts_ft = self._feature_ft(pkt_node, 'end-timestamp-field-type',
                                                     pkt_end_ts_ft)
                    pkt_disc_er_counter_snap_ft = self._feature_ft(pkt_node,
                                                                   'discarded-event-records-counter-snapshot-field-type',
                                                                   pkt_disc_er_counter_snap_ft)
                    pkt_seq_num_ft = self._feature_ft(pkt_node, 'sequence-number-field-type',
                                                      pkt_seq_num_ft)

                # create event record feature field types
                er_node = features_node.get('event-record')
                type_id_ft_prop_name = 'type-id-field-type'

                if er_node is not None:
                    ert_id_ft = self._feature_ft(er_node, type_id_ft_prop_name, ert_id_ft)
                    ert_ts_ft = self._feature_ft(er_node, 'timestamp-field-type', ert_ts_ft)

            erts_prop_name = 'event-record-types'
            ert_count = len(dst_node[erts_prop_name])

            try:
                if ert_id_ft is None and ert_count > 1:
                    raise _ConfigurationParseError(f'`{type_id_ft_prop_name}` property',
                                                   'Event record type ID field type feature is required because data stream type has more than one event record type')

                if isinstance(ert_id_ft, barectf_config._IntegerFieldType):
                    ert_id_int_ft = typing.cast(barectf_config._IntegerFieldType, ert_id_ft)

                    if ert_count > (1 << ert_id_int_ft.size):
                        raise _ConfigurationParseError(f'`{type_id_ft_prop_name}` property',
                                                       f'Field type\'s size ({ert_id_int_ft.size} bits) is too small to accomodate {ert_count} event record types')
            except _ConfigurationParseError as exc:
                exc._append_ctx('`event-record` property')
                _append_error_ctx(exc, '`$features` property')

            pkt_features = barectf_config.DataStreamTypePacketFeatures(pkt_total_size_ft,
                                                                       pkt_content_size_ft,
                                                                       pkt_beginning_ts_ft,
                                                                       pkt_end_ts_ft,
                                                                       pkt_disc_er_counter_snap_ft,
                                                                       pkt_seq_num_ft)
            er_features = barectf_config.DataStreamTypeEventRecordFeatures(ert_id_ft, ert_ts_ft)
            features = barectf_config.DataStreamTypeFeatures(pkt_features, er_features)

            # create packet context (structure) field type extra members
            pkt_ctx_ft_extra_members = None
            prop_name = 'packet-context-field-type-extra-members'
            pkt_ctx_ft_extra_members_node = dst_node.get(prop_name)

            if pkt_ctx_ft_extra_members_node is not None:
                pkt_ctx_ft_extra_members = self._create_struct_ft_members(pkt_ctx_ft_extra_members_node,
                                                                          prop_name)

                # check for illegal packet context field type member names
                reserved_member_names = {
                    'packet_size',
                    'content_size',
                    'timestamp_begin',
                    'timestamp_end',
                    'events_discarded',
                    'packet_seq_num',
                }

                for member_name in pkt_ctx_ft_extra_members:
                    if member_name in reserved_member_names:
                        raise _ConfigurationParseError(f'`{prop_name}` property',
                                                       f'Packet context field type member name `{member_name}` is reserved.')

            # create event record types
            er_header_common_ctx_member_count = Count(0)

            if er_features.type_id_field_type is not None:
                er_header_common_ctx_member_count = Count(er_header_common_ctx_member_count + 1)

            if er_features.timestamp_field_type is not None:
                er_header_common_ctx_member_count = Count(er_header_common_ctx_member_count + 1)

            er_common_ctx_ft_prop_name = 'event-record-common-context-field-type'
            er_common_ctx_ft_node = dst_node.get(er_common_ctx_ft_prop_name)
            er_header_common_ctx_member_count = Count(er_header_common_ctx_member_count + self._total_struct_ft_node_members(er_common_ctx_ft_node))
            erts = set()

            for ert_name, ert_node in dst_node[erts_prop_name].items():
                erts.add(self._create_ert(ert_name, ert_node, er_header_common_ctx_member_count))

            # create data stream type
            return barectf_config.DataStreamType(name, erts, def_clk_type, features,
                                                 pkt_ctx_ft_extra_members,
                                                 self._try_create_struct_ft(dst_node,
                                                                            er_common_ctx_ft_prop_name))
        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, f'Data data stream type `{name}`')

            # satisfy static type checker (never reached)
            raise

    def _clk_type(self, name: str, prop_name: str) -> barectf_config.ClockType:
        clk_type = self._clk_types.get(name)

        if clk_type is None:
            raise _ConfigurationParseError(f'`{prop_name}` property',
                                           f'Clock type `{name}` does not exist')

        return clk_type

    def _create_clk_type(self, name: str, clk_type_node: _MapNode) -> barectf_config.ClockType:
        self._validate_iden(name, '`name` property', 'clock type name')
        clk_type_uuid = None
        uuid_node = clk_type_node.get('uuid')

        if uuid_node is not None:
            clk_type_uuid = uuid.UUID(uuid_node)

        offset_seconds = 0
        offset_cycles = Count(0)
        offset_node = clk_type_node.get('offset')

        if offset_node is not None:
            offset_seconds = offset_node.get('seconds', 0)
            offset_cycles = offset_node.get('cycles', Count(0))

        return barectf_config.ClockType(name, clk_type_node.get('frequency', int(1e9)),
                                        clk_type_uuid, clk_type_node.get('description'),
                                        clk_type_node.get('precision', 0),
                                        barectf_config.ClockTypeOffset(offset_seconds, offset_cycles),
                                        clk_type_node.get('origin-is-unix-epoch', True))

    def _create_clk_types(self):
        self._clk_types = {}

        for clk_type_name, clk_type_node in self._trace_type_node.get('clock-types', {}).items():
            self._clk_types[clk_type_name] = self._create_clk_type(clk_type_name, clk_type_node)

    def _create_trace_type(self):
        try:
            # create clock types (_create_dst() needs them)
            self._create_clk_types()

            # get UUID
            trace_type_uuid = None
            uuid_node = self._trace_type_node.get('uuid')

            if uuid_node is not None:
                if uuid_node == 'auto':
                    trace_type_uuid = uuid.uuid1()
                else:
                    trace_type_uuid = uuid.UUID(uuid_node)

            # create feature field types
            magic_ft = barectf_config.DEFAULT_FIELD_TYPE
            uuid_ft = None
            dst_id_ft = barectf_config.DEFAULT_FIELD_TYPE

            if trace_type_uuid is not None:
                # Trace type has a UUID: initialize UUID field type to
                # a default field type.
                uuid_ft = barectf_config.DEFAULT_FIELD_TYPE

            features_node = self._trace_type_node.get('$features')
            dst_id_ft_prop_name = 'data-stream-type-id-field-type'

            if features_node is not None:
                magic_ft = self._feature_ft(features_node, 'magic-field-type',
                                            magic_ft)
                uuid_ft = self._feature_ft(features_node, 'uuid-field-type', uuid_ft)
                dst_id_ft = self._feature_ft(features_node, dst_id_ft_prop_name, dst_id_ft)

            dsts_prop_name = 'data-stream-types'
            dst_count = len(self._trace_type_node[dsts_prop_name])

            try:
                if dst_id_ft is None and dst_count > 1:
                    raise _ConfigurationParseError(f'`{dst_id_ft_prop_name}` property',
                                                   'Data stream type ID field type feature is required because trace type has more than one data stream type')

                if isinstance(dst_id_ft, barectf_config._FieldType) and dst_count > (1 << dst_id_ft.size):
                    raise _ConfigurationParseError(f'`{dst_id_ft_prop_name}` property',
                                                   f'Field type\'s size ({dst_id_ft.size} bits) is too small to accomodate {dst_count} data stream types')
            except _ConfigurationParseError as exc:
                _append_error_ctx(exc, '`$features` property')

            features = barectf_config.TraceTypeFeatures(magic_ft, uuid_ft, dst_id_ft)

            # create data stream types
            dsts = set()

            for dst_name, dst_node in self._trace_type_node[dsts_prop_name].items():
                dsts.add(self._create_dst(dst_name, dst_node))

            # create trace type
            if self._trace_byte_order_prop_key == 'native-byte-order':
                trace_type_cls = barectf_config.TraceType
            else:
                trace_type_cls = barectf_config.TraceTypeWithUnknownNativeByteOrder

            return trace_type_cls(self._trace_byte_order, dsts, trace_type_uuid, features)
        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, 'Trace type')

    def _create_trace(self):
        try:
            trace_type = self._create_trace_type()
            trace_node = self.config_node['trace']
            env = None
            env_node = trace_node.get('environment')

            if env_node is not None:
                # validate each environment variable name
                for name in env_node:
                    self._validate_iden(name, '`environment` property',
                                        'environment variable name')

                # the node already has the expected structure
                env = barectf_config.TraceEnvironment(env_node)

            return barectf_config.Trace(trace_type, env)

        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, 'Trace')

    def _create_config(self):
        # create trace first
        trace = self._create_trace()

        # find default data stream type, if any
        def_dst = None

        for dst_name, dst_node in self._trace_type_node['data-stream-types'].items():
            prop_name = '$is-default'
            is_default = dst_node.get(prop_name)

            if is_default is True:
                if def_dst is not None:
                    exc = _ConfigurationParseError(f'`{prop_name}` property',
                                                   f'Duplicate default data stream type (`{def_dst.name}`)')
                    exc._append_ctx(f'Data stream type `{dst_name}`')
                    _append_error_ctx(exc, 'Trace type')

                def_dst = trace.type.data_stream_type(dst_name)

        # create clock type C type mapping
        clk_types_node = self._trace_type_node.get('clock-types')
        clk_type_c_types = None

        if clk_types_node is not None:
            clk_type_c_types = collections.OrderedDict()

            for dst in trace.type.data_stream_types:
                if dst.default_clock_type is None:
                    continue

                clk_type_node = clk_types_node[dst.default_clock_type.name]
                c_type = clk_type_node.get('$c-type')

                if c_type is not None:
                    clk_type_c_types[dst.default_clock_type] = c_type

        # create options
        iden_prefix_def = False
        def_dst_name_def = False
        opts_node = self.config_node.get('options')
        iden_prefix = 'barectf_'
        file_name_prefix = 'barectf'

        if opts_node is not None:
            code_gen_opts_node = opts_node.get('code-generation')

            if code_gen_opts_node is not None:
                prefix_node = code_gen_opts_node.get('prefix', 'barectf')

                if type(prefix_node) is str:
                    # automatic prefixes
                    iden_prefix = f'{prefix_node}_'
                    file_name_prefix = prefix_node
                else:
                    iden_prefix = prefix_node['identifier']
                    file_name_prefix = prefix_node['file-name']

                header_opts = code_gen_opts_node.get('header')

                if header_opts is not None:
                    iden_prefix_def = header_opts.get('identifier-prefix-definition', False)
                    def_dst_name_def = header_opts.get('default-data-stream-type-name-definition',
                                                       False)

        header_opts = barectf_config.ConfigurationCodeGenerationHeaderOptions(iden_prefix_def,
                                                                              def_dst_name_def)
        cg_opts = barectf_config.ConfigurationCodeGenerationOptions(iden_prefix, file_name_prefix,
                                                                    def_dst, header_opts,
                                                                    clk_type_c_types)
        opts = barectf_config.ConfigurationOptions(cg_opts)

        # create configuration
        self._config = barectf_config.Configuration(trace, opts)

    # Expands the field type aliases found in the trace type node.
    #
    # This method modifies the trace type node.
    #
    # When this method returns:
    #
    # * Any field type alias is replaced with its full field type
    #   node equivalent.
    #
    # * The `$field-type-aliases` property of the trace type node is
    #   removed.
    def _expand_ft_aliases(self):
        def resolve_ft_alias_from(parent_node: _MapNode, key: str):
            if key not in parent_node:
                return

            if type(parent_node[key]) not in [collections.OrderedDict, str]:
                return

            self._resolve_ft_alias_from(ft_aliases_node, parent_node, key)

        ft_aliases_node = self._trace_type_node['$field-type-aliases']

        # Expand field type aliases within trace, data stream, and event
        # record type nodes.
        features_prop_name = '$features'

        try:
            features_node = self._trace_type_node.get(features_prop_name)

            if features_node is not None:
                try:
                    resolve_ft_alias_from(features_node, 'magic-field-type')
                    resolve_ft_alias_from(features_node, 'uuid-field-type')
                    resolve_ft_alias_from(features_node, 'data-stream-type-id-field-type')
                except _ConfigurationParseError as exc:
                    _append_error_ctx(exc, f'`{features_prop_name}` property')
        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, 'Trace type')

        for dst_name, dst_node in self._trace_type_node['data-stream-types'].items():
            try:
                features_node = dst_node.get(features_prop_name)

                if features_node is not None:
                    try:
                        pkt_prop_name = 'packet'
                        pkt_node = features_node.get(pkt_prop_name)

                        if pkt_node is not None:
                            try:
                                resolve_ft_alias_from(pkt_node, 'total-size-field-type')
                                resolve_ft_alias_from(pkt_node, 'content-size-field-type')
                                resolve_ft_alias_from(pkt_node, 'beginning-timestamp-field-type')
                                resolve_ft_alias_from(pkt_node, 'end-timestamp-field-type')
                                resolve_ft_alias_from(pkt_node,
                                                      'discarded-event-records-counter-snapshot-field-type')
                                resolve_ft_alias_from(pkt_node, 'sequence-number-field-type')
                            except _ConfigurationParseError as exc:
                                _append_error_ctx(exc, f'`{pkt_prop_name}` property')

                        er_prop_name = 'event-record'
                        er_node = features_node.get(er_prop_name)

                        if er_node is not None:
                            try:
                                resolve_ft_alias_from(er_node, 'type-id-field-type')
                                resolve_ft_alias_from(er_node, 'timestamp-field-type')
                            except _ConfigurationParseError as exc:
                                _append_error_ctx(exc, f'`{er_prop_name}` property')
                    except _ConfigurationParseError as exc:
                        _append_error_ctx(exc, f'`{features_prop_name}` property')

                pkt_ctx_ft_extra_members_prop_name = 'packet-context-field-type-extra-members'
                pkt_ctx_ft_extra_members_node = dst_node.get(pkt_ctx_ft_extra_members_prop_name)

                if pkt_ctx_ft_extra_members_node is not None:
                    try:
                        for member_node in pkt_ctx_ft_extra_members_node:
                            member_node = list(member_node.values())[0]
                            resolve_ft_alias_from(member_node, 'field-type')
                    except _ConfigurationParseError as exc:
                        _append_error_ctx(exc, f'`{pkt_ctx_ft_extra_members_prop_name}` property')

                resolve_ft_alias_from(dst_node, 'event-record-common-context-field-type')

                for ert_name, ert_node in dst_node['event-record-types'].items():
                    try:
                        resolve_ft_alias_from(ert_node, 'specific-context-field-type')
                        resolve_ft_alias_from(ert_node, 'payload-field-type')
                    except _ConfigurationParseError as exc:
                        _append_error_ctx(exc, f'Event record type `{ert_name}`')
            except _ConfigurationParseError as exc:
                _append_error_ctx(exc, f'Data stream type `{dst_name}`')

        # remove the (now unneeded) `$field-type-aliases` property
        del self._trace_type_node['$field-type-aliases']

    # Applies field type inheritance to all field type nodes found in
    # the trace type node.
    #
    # This method modifies the trace type node.
    #
    # When this method returns, no field type node has an `$inherit`
    # property.
    def _apply_fts_inheritance(self):
        def apply_ft_inheritance(parent_node: _MapNode, key: str):
            if key not in parent_node:
                return

            if type(parent_node[key]) is not collections.OrderedDict:
                return

            self._apply_ft_inheritance(parent_node, key)

        features_prop_name = '$features'
        features_node = self._trace_type_node.get(features_prop_name)

        if features_node is not None:
            apply_ft_inheritance(features_node, 'magic-field-type')
            apply_ft_inheritance(features_node, 'uuid-field-type')
            apply_ft_inheritance(features_node, 'data-stream-type-id-field-type')

        for dst_node in self._trace_type_node['data-stream-types'].values():
            features_node = dst_node.get(features_prop_name)

            if features_node is not None:
                pkt_node = features_node.get('packet')

                if pkt_node is not None:
                    apply_ft_inheritance(pkt_node, 'total-size-field-type')
                    apply_ft_inheritance(pkt_node, 'content-size-field-type')
                    apply_ft_inheritance(pkt_node, 'beginning-timestamp-field-type')
                    apply_ft_inheritance(pkt_node, 'end-timestamp-field-type')
                    apply_ft_inheritance(pkt_node,
                                         'discarded-event-records-counter-snapshot-field-type')
                    apply_ft_inheritance(pkt_node, 'sequence-number-field-type')

                er_node = features_node.get('event-record')

                if er_node is not None:
                    apply_ft_inheritance(er_node, 'type-id-field-type')
                    apply_ft_inheritance(er_node, 'timestamp-field-type')

            pkt_ctx_ft_extra_members_node = dst_node.get('packet-context-field-type-extra-members')

            if pkt_ctx_ft_extra_members_node is not None:
                for member_node in pkt_ctx_ft_extra_members_node:
                    member_node = list(member_node.values())[0]
                    apply_ft_inheritance(member_node, 'field-type')

            apply_ft_inheritance(dst_node, 'event-record-common-context-field-type')

            for ert_node in dst_node['event-record-types'].values():
                apply_ft_inheritance(ert_node, 'specific-context-field-type')
                apply_ft_inheritance(ert_node, 'payload-field-type')

    # Normalizes structure field type member nodes.
    #
    # A structure field type member node can look like this:
    #
    #     - msg: custom-string
    #
    # which is the equivalent of this:
    #
    #     - msg:
    #         field-type: custom-string
    #
    # This method normalizes form 1 to use form 2.
    def _normalize_struct_ft_member_nodes(self):
        def normalize_members_node(members_node: List[_MapNode]):
            ft_prop_name = 'field-type'

            for member_node in members_node:
                member_name, val_node = list(member_node.items())[0]

                if type(val_node) is str:
                    member_node[member_name] = collections.OrderedDict({
                        ft_prop_name: val_node
                    })

                normalize_struct_ft_member_nodes(member_node[member_name], ft_prop_name)

        def normalize_struct_ft_member_nodes(parent_node: _MapNode, key: str):
            if type(parent_node) is not collections.OrderedDict:
                return

            ft_node = parent_node.get(key)

            if type(ft_node) is not collections.OrderedDict:
                return

            ft_node = typing.cast(collections.OrderedDict, ft_node)
            members_nodes = ft_node.get('members')

            if members_nodes is not None:
                normalize_members_node(members_nodes)

        prop_name = '$field-type-aliases'
        ft_aliases_node = self._trace_type_node.get(prop_name)

        if ft_aliases_node is not None:
            for alias in ft_aliases_node:
                normalize_struct_ft_member_nodes(ft_aliases_node, alias)

        features_prop_name = '$features'
        features_node = self._trace_type_node.get(features_prop_name)

        if features_node is not None:
            normalize_struct_ft_member_nodes(features_node, 'magic-field-type')
            normalize_struct_ft_member_nodes(features_node, 'uuid-field-type')
            normalize_struct_ft_member_nodes(features_node, 'data-stream-type-id-field-type')

        for dst_node in self._trace_type_node['data-stream-types'].values():
            features_node = dst_node.get(features_prop_name)

            if features_node is not None:
                pkt_node = features_node.get('packet')

                if pkt_node is not None:
                    normalize_struct_ft_member_nodes(pkt_node, 'total-size-field-type')
                    normalize_struct_ft_member_nodes(pkt_node, 'content-size-field-type')
                    normalize_struct_ft_member_nodes(pkt_node, 'beginning-timestamp-field-type')
                    normalize_struct_ft_member_nodes(pkt_node, 'end-timestamp-field-type')
                    normalize_struct_ft_member_nodes(pkt_node,
                                                     'discarded-event-records-counter-snapshot-field-type')
                    normalize_struct_ft_member_nodes(pkt_node, 'sequence-number-field-type')

                er_node = features_node.get('event-record')

                if er_node is not None:
                    normalize_struct_ft_member_nodes(er_node, 'type-id-field-type')
                    normalize_struct_ft_member_nodes(er_node, 'timestamp-field-type')

            pkt_ctx_ft_extra_members_node = dst_node.get('packet-context-field-type-extra-members')

            if pkt_ctx_ft_extra_members_node is not None:
                normalize_members_node(pkt_ctx_ft_extra_members_node)

            normalize_struct_ft_member_nodes(dst_node, 'event-record-common-context-field-type')

            for ert_node in dst_node['event-record-types'].values():
                normalize_struct_ft_member_nodes(ert_node, 'specific-context-field-type')
                normalize_struct_ft_member_nodes(ert_node, 'payload-field-type')

    # Calls _expand_ft_aliases() and _apply_fts_inheritance() if the
    # trace type node has a `$field-type-aliases` property.
    def _expand_fts(self):
        # Make sure that the current configuration node is valid
        # considering field types are not expanded yet.
        self._schema_validator.validate(self.config_node,
                                        'config/3/config-pre-field-type-expansion')

        prop_name = '$field-type-aliases'
        ft_aliases_node = self._trace_type_node.get(prop_name)

        if ft_aliases_node is None:
            # If there's no `'$field-type-aliases'` node, then there's
            # no field type aliases and therefore no possible
            # inheritance.
            if prop_name in self._trace_type_node:
                del self._trace_type_node[prop_name]

            return

        # normalize structure field type member nodes
        self._normalize_struct_ft_member_nodes()

        # first, expand field type aliases
        self._expand_ft_aliases()

        # next, apply inheritance to create effective field type nodes
        self._apply_fts_inheritance()

    # Substitute the event record type node log level aliases with their
    # numeric equivalents.
    #
    # Removes the `$log-level-aliases` property of the trace type node.
    def _sub_log_level_aliases(self):
        # Make sure that the current configuration node is valid
        # considering log level aliases are not substituted yet.
        self._schema_validator.validate(self.config_node,
                                        'config/3/config-pre-log-level-alias-sub')

        log_level_aliases_prop_name = '$log-level-aliases'
        log_level_aliases_node = self._trace_type_node.get(log_level_aliases_prop_name)

        if log_level_aliases_prop_name in self._trace_type_node:
            del self._trace_type_node[log_level_aliases_prop_name]

        if log_level_aliases_node is None:
            # no log level aliases
            return

        # substitute log level aliases
        for dst_name, dst_node in self._trace_type_node['data-stream-types'].items():
            try:
                for ert_name, ert_node in dst_node['event-record-types'].items():
                    try:
                        prop_name = 'log-level'
                        ll_node = ert_node.get(prop_name)

                        if ll_node is None:
                            continue

                        if type(ll_node) is str:
                            if ll_node not in log_level_aliases_node:
                                raise _ConfigurationParseError(f'`{prop_name}` property',
                                                               f'Log level alias `{ll_node}` does not exist')

                            ert_node[prop_name] = log_level_aliases_node[ll_node]
                    except _ConfigurationParseError as exc:
                        _append_error_ctx(exc, f'Event record type `{ert_name}`')
            except _ConfigurationParseError as exc:
                _append_error_ctx(exc, f'Data stream type `{dst_name}`')

    # Generator of parent node and key pairs for all the nodes,
    # recursively, of `node`.
    #
    # It is safe to delete a yielded node during the iteration.
    @staticmethod
    def _props(node: Any) -> Iterable[Tuple[Any, str]]:
        if type(node) is collections.OrderedDict:
            for key in list(node):
                yield from _Parser._props(node[key])
                yield node, key
        elif type(node) is list:
            for item_node in node:
                yield from _Parser._props(item_node)

    def _trace_type_props(self) -> Iterable[Tuple[Any, str]]:
        yield from _Parser._props(self.config_node['trace']['type'])

    def _set_trace_byte_order_prop_key(self):
        if 'native-byte-order' in self._trace_type_node:
            self._trace_byte_order_prop_key = 'native-byte-order'
        else:
            self._trace_byte_order_prop_key = 'trace-byte-order'

    # Normalize the properties of the configuration node.
    #
    # This method, for each property of the trace type node:
    #
    # 1. Removes it if it's `None` (means default).
    #
    # 2. Chooses a specific `class` property value.
    #
    # 3. Chooses a specific trace byte order property value.
    #
    # 4. Chooses a specific `preferred-display-base` property value.
    #
    # This method also applies 1. to the trace node's `environment`
    # property.
    def _normalize_props(self):
        def normalize_byte_order_prop(parent_node: _MapNode, key: str):
            node = parent_node[key]

            if node in ['be', 'big']:
                parent_node[key] = 'big-endian'
            elif node in ['le', 'little']:
                parent_node[key] = 'little-endian'

        trace_node = self.config_node['trace']
        normalize_byte_order_prop(self._trace_type_node, self._trace_byte_order_prop_key)

        for parent_node, key in self._trace_type_props():
            node = parent_node[key]

            if node is None:
                # a `None` property is equivalent to not having it
                del parent_node[key]
                continue

            if key == 'class' and type(node) is str:
                # field type class aliases
                if node in ['uint', 'unsigned-int']:
                    parent_node[key] = 'unsigned-integer'
                elif node in ['sint', 'signed-int']:
                    parent_node[key] = 'signed-integer'
                elif node in ['uenum', 'unsigned-enum']:
                    parent_node[key] = 'unsigned-enumeration'
                elif node in ['senum', 'signed-enum']:
                    parent_node[key] = 'signed-enumeration'
                elif node == 'str':
                    parent_node[key] = 'string'
                elif node == 'struct':
                    parent_node[key] = 'structure'
            elif key == 'preferred-display-base' and type(node) is str:
                # display base aliases
                if node == 'bin':
                    parent_node[key] = 'binary'
                elif node == 'oct':
                    parent_node[key] = 'octal'
                elif node == 'dec':
                    parent_node[key] = 'decimal'
                elif node == 'hex':
                    parent_node[key] = 'hexadecimal'

        prop_name = 'environment'

        if prop_name in trace_node:
            node = trace_node[prop_name]

            if node is None:
                del trace_node[prop_name]

    # Sets the parser's trace byte order.
    def _set_trace_byte_order(self):
        self._trace_byte_order_node = self._trace_type_node[self._trace_byte_order_prop_key]
        self._trace_byte_order = self._byte_order_from_node(self._trace_byte_order_node)

    # Processes the inclusions of the event record type node
    # `ert_node`, returning the effective node.
    def _process_ert_node_include(self, ert_node: _MapNode) -> _MapNode:
        # Make sure the event record type node is valid for the
        # inclusion processing stage.
        self._schema_validator.validate(ert_node, 'config/3/ert-pre-include')

        # process inclusions
        return self._process_node_include(ert_node, self._process_ert_node_include)

    # Processes the inclusions of the data stream type node `dst_node`,
    # returning the effective node.
    def _process_dst_node_include(self, dst_node: _MapNode) -> _MapNode:
        def process_children_include(dst_node: _MapNode):
            prop_name = 'event-record-types'

            if prop_name in dst_node:
                erts_node = dst_node[prop_name]

                for key in list(erts_node):
                    erts_node[key] = self._process_ert_node_include(erts_node[key])

        # Make sure the data stream type node is valid for the inclusion
        # processing stage.
        self._schema_validator.validate(dst_node, 'config/3/dst-pre-include')

        # process inclusions
        return self._process_node_include(dst_node, self._process_dst_node_include,
                                          process_children_include)

    # Processes the inclusions of the clock type node `clk_type_node`,
    # returning the effective node.
    def _process_clk_type_node_include(self, clk_type_node: _MapNode) -> _MapNode:
        # Make sure the clock type node is valid for the inclusion
        # processing stage.
        self._schema_validator.validate(clk_type_node, 'config/3/clock-type-pre-include')

        # process inclusions
        return self._process_node_include(clk_type_node, self._process_clk_type_node_include)

    # Processes the inclusions of the trace type node `trace_type_node`,
    # returning the effective node.
    def _process_trace_type_node_include(self, trace_type_node: _MapNode) -> _MapNode:
        def process_children_include(trace_type_node: _MapNode):
            prop_name = 'clock-types'

            if prop_name in trace_type_node:
                clk_types_node = trace_type_node[prop_name]

                for key in list(clk_types_node):
                    clk_types_node[key] = self._process_clk_type_node_include(clk_types_node[key])

            prop_name = 'data-stream-types'

            if prop_name in trace_type_node:
                dsts_node = trace_type_node[prop_name]

                for key in list(dsts_node):
                    dsts_node[key] = self._process_dst_node_include(dsts_node[key])

        # Make sure the trace type node is valid for the inclusion
        # processing stage.
        self._schema_validator.validate(trace_type_node, 'config/3/trace-type-pre-include')

        # process inclusions
        return self._process_node_include(trace_type_node, self._process_trace_type_node_include,
                                          process_children_include)

    # Processes the inclusions of the trace node `trace_node`, returning
    # the effective node.
    def _process_trace_node_include(self, trace_node: _MapNode) -> _MapNode:
        def process_children_include(trace_node: _MapNode):
            prop_name = 'type'

            if prop_name in trace_node:
                trace_node[prop_name] = self._process_trace_type_node_include(trace_node[prop_name])

        # Make sure the trace node is valid for the inclusion processing
        # stage.
        self._schema_validator.validate(trace_node, 'config/3/trace-pre-include')

        # process inclusions
        return self._process_node_include(trace_node, self._process_trace_node_include,
                                          process_children_include)

    # Processes the inclusions of the configuration node, modifying it
    # during the process.
    def _process_config_includes(self):
        # Process inclusions in this order:
        #
        # 1. Clock type node and event record type nodes (the order
        #    between those is not important).
        #
        # 2. Data stream type nodes.
        #
        # 3. Trace type node.
        #
        # 4. Trace node.
        #
        # This is because:
        #
        # * A trace node can include a trace type node, clock type
        #   nodes, data stream type nodes, and event record type nodes.
        #
        # * A trace type node can include clock type nodes, data stream
        #   type nodes, and event record type nodes.
        #
        # * A data stream type node can include event record type nodes.
        #
        # First, make sure the configuration node itself is valid for
        # the inclusion processing stage.
        self._schema_validator.validate(self.config_node, 'config/3/config-pre-include')

        # Process trace node inclusions.
        #
        # self._process_trace_node_include() returns a new (or the same)
        # trace node without any `$include` property in it, recursively.
        self.config_node['trace'] = self._process_trace_node_include(self.config_node['trace'])

    def _parse(self):
        # process configuration node inclusions
        self._process_config_includes()

        # Expand field type nodes.
        #
        # This process:
        #
        # 1. Replaces field type aliases with "effective" field type
        #    nodes, recursively.
        #
        #    After this step, the `$field-type-aliases` property of the
        #    trace type node is gone.
        #
        # 2. Applies inheritance, following the `$inherit` properties.
        #
        #    After this step, field type nodes do not contain `$inherit`
        #    properties.
        #
        # This is done blindly, in that the process _doesn't_ validate
        # field type nodes at this point.
        self._expand_fts()

        # Substitute log level aliases.
        #
        # This process:
        #
        # 1. Replaces log level aliases in event record type nodes with
        #    their numeric equivalents as found in the
        #    `$log-level-aliases` property of the trace type node.
        #
        # 2. Removes the `$log-level-aliases` property from the trace
        #    type node.
        self._sub_log_level_aliases()

        # At this point, the configuration node must be valid as an
        # effective configuration node.
        self._schema_validator.validate(self.config_node, 'config/3/config')

        # Set the trace byte order property key.
        self._set_trace_byte_order_prop_key()

        # Normalize properties.
        #
        # This process removes `None` properties and chooses specific
        # enumerators when aliases exist (for example, `big-endian`
        # instead of `be`).
        #
        # The goal of this is that, if the user then gets this parser's
        # `config_node` property, it has a normal and very readable
        # form.
        #
        # It also makes _create_config() easier to implement because it
        # doesn't need to check for `None` nodes or enumerator aliases.
        self._normalize_props()

        # Set the trace byte order.
        self._set_trace_byte_order()

        # Create a barectf configuration object from the configuration
        # node.
        self._create_config()

    @property
    def config(self) -> barectf_config.Configuration:
        return self._config

    @property
    def config_node(self) -> _MapNode:
        return typing.cast(barectf_config_parse_common._ConfigNodeV3, self._root_node).config_node



# The MIT License (MIT)
#
# Copyright (c) 2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import jinja2
import barectf.config as barectf_config
import barectf.version as barectf_version
from barectf.typing import Count
from typing import Callable, Optional, Any, Mapping


def _filt_indent_tab(text: str, count: Count = Count(1), indent_first: bool = False) -> str:
    in_lines = text.split('\n')
    out_lines = []
    tab = '\t'

    for index, in_line in enumerate(in_lines):
        in_line = in_line.rstrip()

        if len(in_line) == 0 or (index == 0 and not indent_first):
            out_lines.append(in_line)
            continue

        out_lines.append(f'{tab * count}{in_line}')

    return '\n'.join(out_lines)


def _filt_escape_dq(text: str) -> str:
    return text.replace('\\', '\\\\').replace('"', '\\"')


_Filter = Callable[..., Any]
_Filters = Mapping[str, _Filter]
_Test = Callable[..., bool]
_Tests = Mapping[str, _Test]


class _Template:
    def __init__(self, name: str, is_file_template: bool = False,
                 cfg: Optional[barectf_config.Configuration] = None,
                 filters: Optional[_Filters] = None, tests: Optional[_Tests] = None):
        env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True,
                                 loader=jinja2.PackageLoader('barectf', 'templates'))
        env.globals.update({
            'cfg': cfg,
            'barectf_config': barectf_config,
            'barectf_version': barectf_version,
        })

        env.filters['indent_tab'] = _filt_indent_tab
        env.filters['escape_dq'] = _filt_escape_dq

        if filters is not None:
            env.filters.update(filters)

        if tests is not None:
            env.tests.update(tests)

        self._templ = env.get_template(name)
        self._is_file_template = is_file_template

    def render(self, **kwargs) -> str:
        text = self._templ.render(**kwargs)

        if self._is_file_template:
            text = text.strip() + '\n'

        return text



# The MIT License (MIT)
#
# Copyright (c) 2015-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from barectf.config_parse_common import _ConfigurationParseError
from barectf.config_parse_common import _append_error_ctx
import barectf.config_parse_common as config_parse_common
from barectf.config_parse_common import _MapNode
import collections
import copy
from barectf.typing import VersionNumber, _OptStr
from typing import Optional, List, Dict, TextIO, Union, Callable
import typing


def _del_prop_if_exists(node: _MapNode, prop_name: str):
    if prop_name in node:
        del node[prop_name]


def _rename_prop(node: _MapNode, old_prop_name: str, new_prop_name: str):
    if old_prop_name in node:
        node[new_prop_name] = node[old_prop_name]
        del node[old_prop_name]


def _copy_prop_if_exists(dst_node: _MapNode, src_node: _MapNode, src_prop_name: str,
                         dst_prop_name: _OptStr = None):
    if dst_prop_name is None:
        dst_prop_name = src_prop_name

    if src_prop_name in src_node:
        dst_node[dst_prop_name] = copy.deepcopy(src_node[src_prop_name])


# A barectf 2 YAML configuration parser.
#
# The only purpose of such a parser is to transform the passed root
# configuration node so that it's a valid barectf 3 configuration node.
#
# The parser's `config_node` property is the equivalent barectf 3
# configuration node.
#
# See the comments of _parse() for more implementation details about the
# parsing stages and general strategy.
class _Parser(config_parse_common._Parser):
    # Builds a barectf 2 YAML configuration parser and parses the root
    # configuration node `node` (already loaded from the file-like
    # object `root_file`).
    def __init__(self, root_file: TextIO, node: _MapNode, with_pkg_include_dir: bool,
                 include_dirs: Optional[List[str]], ignore_include_not_found: bool):
        super().__init__(root_file, node, with_pkg_include_dir, include_dirs,
                         ignore_include_not_found, VersionNumber(2))
        self._ft_cls_name_to_conv_method: Dict[str, Callable[[_MapNode], _MapNode]] = {
            'int': self._conv_int_ft_node,
            'integer': self._conv_int_ft_node,
            'enum': self._conv_enum_ft_node,
            'enumeration': self._conv_enum_ft_node,
            'flt': self._conv_real_ft_node,
            'float': self._conv_real_ft_node,
            'floating-point': self._conv_real_ft_node,
            'str': self._conv_string_ft_node,
            'string': self._conv_string_ft_node,
            'array': self._conv_array_ft_node,
            'struct': self._conv_struct_ft_node,
            'structure': self._conv_struct_ft_node,
        }
        self._parse()

    # Converts a v2 field type node to a v3 field type node and returns
    # it.
    def _conv_ft_node(self, v2_ft_node: _MapNode) -> _MapNode:
        assert 'class' in v2_ft_node
        cls = v2_ft_node['class']
        assert cls in self._ft_cls_name_to_conv_method
        return self._ft_cls_name_to_conv_method[cls](v2_ft_node)

    def _conv_ft_node_if_exists(self, v2_parent_node: Optional[_MapNode], key: str) -> Optional[_MapNode]:
        if v2_parent_node is None:
            return None

        if key not in v2_parent_node:
            return None

        return self._conv_ft_node(v2_parent_node[key])

    # Converts a v2 integer field type node to a v3 integer field type
    # node and returns it.
    def _conv_int_ft_node(self, v2_ft_node: _MapNode) -> _MapNode:
        # copy v2 integer field type node
        v3_ft_node = copy.deepcopy(v2_ft_node)

        # signedness depends on the class, not a property
        cls_name = 'uint'
        prop_name = 'signed'
        is_signed_node = v3_ft_node.get(prop_name)

        if is_signed_node is True:
            cls_name = 'sint'

        v3_ft_node['class'] = cls_name
        _del_prop_if_exists(v3_ft_node, prop_name)

        # rename `align` property to `alignment`
        _rename_prop(v3_ft_node, 'align', 'alignment')

        # rename `base` property to `preferred-display-base`
        _rename_prop(v3_ft_node, 'base', 'preferred-display-base')

        # remove `encoding` property
        _del_prop_if_exists(v3_ft_node, 'encoding')

        # remove `byte-order` property (the equivalent barectf 3
        # configuration property is named `trace-byte-order`)
        _del_prop_if_exists(v3_ft_node, 'byte-order')

        # remove `property-mappings` property
        _del_prop_if_exists(v3_ft_node, 'property-mappings')

        return v3_ft_node

    # Converts a v2 enumeration field type node to a v3 enumeration
    # field type node and returns it.
    def _conv_enum_ft_node(self, v2_ft_node: _MapNode) -> _MapNode:
        # An enumeration field type _is_ an integer field type, so use a
        # copy of the converted v2 value field type node.
        v3_ft_node = copy.deepcopy(self._conv_ft_node(v2_ft_node['value-type']))

        # transform class name accordingly
        prop_name = 'class'
        cls_name = 'uenum'

        if v3_ft_node[prop_name] == 'sint':
            cls_name = 'senum'

        v3_ft_node[prop_name] = cls_name

        # convert members to mappings
        prop_name = 'members'
        members_node = v2_ft_node.get(prop_name)

        if members_node is not None:
            mappings_node: _MapNode = collections.OrderedDict()
            cur = 0

            for member_node in members_node:
                v3_value_node: Union[int, List[int]]

                if type(member_node) is str:
                    label = member_node
                    v3_value_node = cur
                    cur += 1
                else:
                    assert type(member_node) is collections.OrderedDict
                    label = member_node['label']
                    v2_value_node = member_node['value']

                    if type(v2_value_node) is int:
                        cur = v2_value_node + 1
                        v3_value_node = v2_value_node
                    else:
                        assert type(v2_value_node) is list
                        assert len(v2_value_node) == 2
                        v3_value_node = list(v2_value_node)
                        cur = v2_value_node[1] + 1

                if label not in mappings_node:
                    mappings_node[label] = []

                mappings_node[label].append(v3_value_node)

            v3_ft_node['mappings'] = mappings_node

        return v3_ft_node

    # Converts a v2 real field type node to a v3 real field type node
    # and returns it.
    def _conv_real_ft_node(self, v2_ft_node: _MapNode) -> _MapNode:
        # copy v2 real field type node
        v3_ft_node = copy.deepcopy(v2_ft_node)

        # set class to `real`
        v3_ft_node['class'] = 'real'

        # rename `align` property to `alignment`
        _rename_prop(v3_ft_node, 'align', 'alignment')

        # set `size` property to a single integer (total size, in bits)
        prop_name = 'size'
        v3_ft_node[prop_name] = v3_ft_node[prop_name]['exp'] + v3_ft_node[prop_name]['mant']

        return v3_ft_node

    # Converts a v2 string field type node to a v3 string field type
    # node and returns it.
    def _conv_string_ft_node(self, v2_ft_node: _MapNode) -> _MapNode:
        # copy v2 string field type node
        v3_ft_node = copy.deepcopy(v2_ft_node)

        # remove `encoding` property
        _del_prop_if_exists(v3_ft_node, 'encoding')

        return v3_ft_node

    # Converts a v2 array field type node to a v3 (static) array field
    # type node and returns it.
    def _conv_array_ft_node(self, v2_ft_node: _MapNode) -> _MapNode:
        # class renamed to `static-array` or `dynamic-array`
        is_dynamic = v2_ft_node['length'] == 'dynamic'
        array_type = 'dynamic' if is_dynamic else 'static'
        v3_ft_node: _MapNode = collections.OrderedDict({'class': f'{array_type}-array'})

        # copy `length` property if it's a static array field type
        if not is_dynamic:
            _copy_prop_if_exists(v3_ft_node, v2_ft_node, 'length')

        # convert element field type
        v3_ft_node['element-field-type'] = self._conv_ft_node(v2_ft_node['element-type'])

        return v3_ft_node

    # Converts a v2 structure field type node to a v3 structure field
    # type node and returns it.
    def _conv_struct_ft_node(self, v2_ft_node: _MapNode) -> _MapNode:
        # Create fresh v3 structure field type node, reusing the class
        # of `v2_ft_node`.
        v3_ft_node = collections.OrderedDict({'class': v2_ft_node['class']})

        # rename `min-align` property to `minimum-alignment`
        _copy_prop_if_exists(v3_ft_node, v2_ft_node, 'min-align', 'minimum-alignment')

        # convert fields to members
        prop_name = 'fields'

        if prop_name in v2_ft_node:
            members_node = []

            for member_name, v2_member_ft_node in v2_ft_node[prop_name].items():
                members_node.append(collections.OrderedDict({
                    member_name: collections.OrderedDict({
                        'field-type': self._conv_ft_node(v2_member_ft_node)
                    })
                }))

            v3_ft_node['members'] = members_node

        return v3_ft_node

    # Converts a v2 clock type node to a v3 clock type node and returns
    # it.
    def _conv_clk_type_node(self, v2_clk_type_node: _MapNode) -> _MapNode:
        # copy v2 clock type node
        v3_clk_type_node = copy.deepcopy(v2_clk_type_node)

        # rename `freq` property to `frequency`
        _rename_prop(v3_clk_type_node, 'freq', 'frequency')

        # rename `error-cycles` property to `precision`
        _rename_prop(v3_clk_type_node, 'error-cycles', 'precision')

        # rename `absolute` property to `origin-is-unix-epoch`
        _rename_prop(v3_clk_type_node, 'absolute', 'origin-is-unix-epoch')

        # rename `$return-ctype`/`return-ctype` property to `$c-type`
        new_prop_name = '$c-type'
        _rename_prop(v3_clk_type_node, 'return-ctype', new_prop_name)
        _rename_prop(v3_clk_type_node, '$return-ctype', new_prop_name)

        return v3_clk_type_node

    # Converts a v2 event record type node to a v3 event record type
    # node and returns it.
    def _conv_ert_node(self, v2_ert_node: _MapNode) -> _MapNode:
        # create empty v3 event record type node
        v3_ert_node: _MapNode = collections.OrderedDict()

        # copy `log-level` property
        _copy_prop_if_exists(v3_ert_node, v2_ert_node, 'log-level')

        # convert specific context field type node
        v2_ft_node = v2_ert_node.get('context-type')

        if v2_ft_node is not None:
            v3_ert_node['specific-context-field-type'] = self._conv_ft_node(v2_ft_node)

        # convert payload field type node
        v2_ft_node = v2_ert_node.get('payload-type')

        if v2_ft_node is not None:
            v3_ert_node['payload-field-type'] = self._conv_ft_node(v2_ft_node)

        return v3_ert_node

    @staticmethod
    def _set_v3_feature_ft_if_exists(v3_features_node: _MapNode, key: str,
                                     node: Union[Optional[_MapNode], bool]):
        val = node

        if val is None:
            val = False

        v3_features_node[key] = val

    # Converts a v2 data stream type node to a v3 data stream type node
    # and returns it.
    def _conv_dst_node(self, v2_dst_node: _MapNode) -> _MapNode:
        # This function creates a v3 data stream type features node from
        # the packet context and event record header field type nodes of
        # a v2 data stream type node.
        def v3_features_node_from_v2_ft_nodes(v2_pkt_ctx_ft_fields_node: _MapNode,
                                              v2_er_header_ft_fields_node: Optional[_MapNode]) -> _MapNode:
            if v2_er_header_ft_fields_node is None:
                v2_er_header_ft_fields_node = collections.OrderedDict()

            v3_pkt_total_size_ft_node = self._conv_ft_node(v2_pkt_ctx_ft_fields_node['packet_size'])
            v3_pkt_content_size_ft_node = self._conv_ft_node(v2_pkt_ctx_ft_fields_node['content_size'])
            v3_pkt_beg_ts_ft_node = self._conv_ft_node_if_exists(v2_pkt_ctx_ft_fields_node,
                                                                 'timestamp_begin')
            v3_pkt_end_ts_ft_node = self._conv_ft_node_if_exists(v2_pkt_ctx_ft_fields_node,
                                                                 'timestamp_end')
            v3_pkt_disc_er_counter_snap_ft_node = self._conv_ft_node_if_exists(v2_pkt_ctx_ft_fields_node,
                                                                               'events_discarded')
            v3_ert_id_ft_node = self._conv_ft_node_if_exists(v2_er_header_ft_fields_node, 'id')
            v3_er_ts_ft_node = self._conv_ft_node_if_exists(v2_er_header_ft_fields_node,
                                                            'timestamp')
            v3_features_node: _MapNode = collections.OrderedDict()
            v3_pkt_node: _MapNode = collections.OrderedDict()
            v3_er_node: _MapNode = collections.OrderedDict()
            v3_pkt_node['total-size-field-type'] = v3_pkt_total_size_ft_node
            v3_pkt_node['content-size-field-type'] = v3_pkt_content_size_ft_node
            self._set_v3_feature_ft_if_exists(v3_pkt_node, 'beginning-timestamp-field-type',
                                              v3_pkt_beg_ts_ft_node)
            self._set_v3_feature_ft_if_exists(v3_pkt_node, 'end-timestamp-field-type',
                                              v3_pkt_end_ts_ft_node)
            self._set_v3_feature_ft_if_exists(v3_pkt_node,
                                              'discarded-event-records-counter-snapshot-field-type',
                                              v3_pkt_disc_er_counter_snap_ft_node)
            self._set_v3_feature_ft_if_exists(v3_er_node, 'type-id-field-type', v3_ert_id_ft_node)
            self._set_v3_feature_ft_if_exists(v3_er_node, 'timestamp-field-type', v3_er_ts_ft_node)
            v3_features_node['packet'] = v3_pkt_node
            v3_features_node['event-record'] = v3_er_node
            return v3_features_node

        def clk_type_name_from_v2_int_ft_node(v2_int_ft_node: Optional[_MapNode]) -> _OptStr:
            if v2_int_ft_node is None:
                return None

            assert v2_int_ft_node['class'] in ('int', 'integer')
            prop_mappings_node = v2_int_ft_node.get('property-mappings')

            if prop_mappings_node is not None and len(prop_mappings_node) > 0:
                return prop_mappings_node[0]['name']

            return None

        # create empty v3 data stream type node
        v3_dst_node: _MapNode = collections.OrderedDict()

        # rename `$default` property to `$is-default`
        _copy_prop_if_exists(v3_dst_node, v2_dst_node, '$default', '$is-default')

        # set default clock type node
        pct_prop_name = 'packet-context-type'
        v2_pkt_ctx_ft_fields_node = v2_dst_node[pct_prop_name]['fields']
        eht_prop_name = 'event-header-type'
        v2_er_header_ft_fields_node = None
        v2_er_header_ft_node = v2_dst_node.get(eht_prop_name)

        if v2_er_header_ft_node is not None:
            v2_er_header_ft_fields_node = v2_er_header_ft_node['fields']

        def_clk_type_name = None

        try:
            ts_begin_prop_name = 'timestamp_begin'
            ts_begin_clk_type_name = clk_type_name_from_v2_int_ft_node(v2_pkt_ctx_ft_fields_node.get(ts_begin_prop_name))
            ts_end_prop_name = 'timestamp_end'
            ts_end_clk_type_name = clk_type_name_from_v2_int_ft_node(v2_pkt_ctx_ft_fields_node.get(ts_end_prop_name))

            if ts_begin_clk_type_name is not None and ts_end_clk_type_name is not None:
                if ts_begin_clk_type_name != ts_end_clk_type_name:
                    raise _ConfigurationParseError(f'`{ts_begin_prop_name}`/`{ts_end_prop_name}` properties',
                                                   'Field types are not mapped to the same clock type')
        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, f'`{pct_prop_name}` property')

        try:
            if def_clk_type_name is None and v2_er_header_ft_fields_node is not None:
                def_clk_type_name = clk_type_name_from_v2_int_ft_node(v2_er_header_ft_fields_node.get('timestamp'))

            if def_clk_type_name is None and ts_begin_clk_type_name is not None:
                def_clk_type_name = ts_begin_clk_type_name

            if def_clk_type_name is None and ts_end_clk_type_name is not None:
                def_clk_type_name = ts_end_clk_type_name
        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, f'`{eht_prop_name}` property')

        if def_clk_type_name is not None:
            v3_dst_node['$default-clock-type-name'] = def_clk_type_name

        # set features node
        v3_dst_node['$features'] = v3_features_node_from_v2_ft_nodes(v2_pkt_ctx_ft_fields_node,
                                                                     v2_er_header_ft_fields_node)

        # set extra packet context field type members node
        pkt_ctx_ft_extra_members = []
        ctf_member_names = [
            'packet_size',
            'content_size',
            'timestamp_begin',
            'timestamp_end',
            'events_discarded',
            'packet_seq_num',
        ]

        for member_name, v2_ft_node in v2_pkt_ctx_ft_fields_node.items():
            if member_name in ctf_member_names:
                continue

            pkt_ctx_ft_extra_members.append(collections.OrderedDict({
                member_name: collections.OrderedDict({
                    'field-type': self._conv_ft_node(v2_ft_node)
                })
            }))

        if len(pkt_ctx_ft_extra_members) > 0:
            v3_dst_node['packet-context-field-type-extra-members'] = pkt_ctx_ft_extra_members

        # convert event record common context field type node
        v2_ft_node = v2_dst_node.get('event-context-type')

        if v2_ft_node is not None:
            v3_dst_node['event-record-common-context-field-type'] = self._conv_ft_node(v2_ft_node)

        # convert event record type nodes
        v3_erts_node = collections.OrderedDict()

        for ert_name, v2_ert_node in v2_dst_node['events'].items():
            try:
                v3_erts_node[ert_name] = self._conv_ert_node(v2_ert_node)
            except _ConfigurationParseError as exc:
                _append_error_ctx(exc, f'Event record type `{ert_name}`')

        v3_dst_node['event-record-types'] = v3_erts_node

        return v3_dst_node

    # Converts a v2 metadata node to a v3 trace node and returns it.
    def _conv_meta_node(self, v2_meta_node: _MapNode) -> _MapNode:
        def v3_features_node_from_v2_ft_node(v2_pkt_header_ft_node: Optional[_MapNode]) -> _MapNode:
            def set_if_exists(key, node):
                return self._set_v3_feature_ft_if_exists(v3_features_node, key, node)

            v2_pkt_header_ft_fields_node = collections.OrderedDict()

            if v2_pkt_header_ft_node is not None:
                v2_pkt_header_ft_fields_node = v2_pkt_header_ft_node['fields']

            v3_magic_ft_node = self._conv_ft_node_if_exists(v2_pkt_header_ft_fields_node, 'magic')
            v3_uuid_ft_node = self._conv_ft_node_if_exists(v2_pkt_header_ft_fields_node, 'uuid')
            v3_dst_id_ft_node = self._conv_ft_node_if_exists(v2_pkt_header_ft_fields_node,
                                                             'stream_id')
            v3_features_node: _MapNode = collections.OrderedDict()
            set_if_exists('magic-field-type', v3_magic_ft_node)
            set_if_exists('uuid-field-type', v3_uuid_ft_node)
            set_if_exists('data-stream-type-id-field-type', v3_dst_id_ft_node)
            return v3_features_node

        v3_trace_node: _MapNode = collections.OrderedDict()
        v3_trace_type_node: _MapNode = collections.OrderedDict()
        v2_trace_node = v2_meta_node['trace']

        # copy `byte-order` property as `trace-byte-order` property
        _copy_prop_if_exists(v3_trace_type_node, v2_trace_node, 'byte-order', 'trace-byte-order')

        # copy `uuid` property
        _copy_prop_if_exists(v3_trace_type_node, v2_trace_node, 'uuid')

        # copy `$log-levels`/`log-levels` property
        new_prop_name = '$log-level-aliases'
        _copy_prop_if_exists(v3_trace_type_node, v2_meta_node, 'log-levels', new_prop_name)
        _copy_prop_if_exists(v3_trace_type_node, v2_meta_node, '$log-levels', new_prop_name)

        # copy `clocks` property, converting clock type nodes
        v2_clk_types_node = v2_meta_node.get('clocks')

        if v2_clk_types_node is not None:
            v3_clk_types_node = collections.OrderedDict()

            for name, v2_clk_type_node in v2_clk_types_node.items():
                v3_clk_types_node[name] = self._conv_clk_type_node(v2_clk_type_node)

            v3_trace_type_node['clock-types'] = v3_clk_types_node

        # set features node
        v2_pkt_header_ft_node = v2_trace_node.get('packet-header-type')
        v3_trace_type_node['$features'] = v3_features_node_from_v2_ft_node(v2_pkt_header_ft_node)

        # convert data stream type nodes
        v3_dsts_node = collections.OrderedDict()

        for dst_name, v2_dst_node in v2_meta_node['streams'].items():
            try:
                v3_dsts_node[dst_name] = self._conv_dst_node(v2_dst_node)
            except _ConfigurationParseError as exc:
                _append_error_ctx(exc, f'Data stream type `{dst_name}`')

        v3_trace_type_node['data-stream-types'] = v3_dsts_node

        # If `v2_meta_node` has a `$default-stream` property, find the
        # corresponding v3 data stream type node and set its
        # `$is-default` property to `True`.
        prop_name = '$default-stream'
        v2_def_dst_node = v2_meta_node.get(prop_name)

        if v2_def_dst_node is not None:
            found = False

            for dst_name, v3_dst_node in v3_dsts_node.items():
                if dst_name == v2_def_dst_node:
                    v3_dst_node['$is-default'] = True
                    found = True
                    break

            if not found:
                raise _ConfigurationParseError(f'`{prop_name}` property',
                                               f'Data stream type `{v2_def_dst_node}` does not exist')

        # set environment node
        v2_env_node = v2_meta_node.get('env')

        if v2_env_node is not None:
            v3_trace_node['environment'] = copy.deepcopy(v2_env_node)

        # set v3 trace node's type node
        v3_trace_node['type'] = v3_trace_type_node

        return v3_trace_node

    # Transforms the root configuration node into a valid v3
    # configuration node.
    def _transform_config_node(self):
        # remove the `version` property
        del self._root_node['version']

        # relocate prefix and option nodes
        prefix_prop_name = 'prefix'
        v2_prefix_node = self._root_node.get(prefix_prop_name, 'barectf_')
        _del_prop_if_exists(self._root_node, prefix_prop_name)
        opt_prop_name = 'options'
        v2_options_node = self._root_node.get(opt_prop_name)
        _del_prop_if_exists(self._root_node, opt_prop_name)
        code_gen_node = collections.OrderedDict()
        v3_prefixes = config_parse_common._v3_prefixes_from_v2_prefix(v2_prefix_node)
        v3_prefix_node = collections.OrderedDict([
            ('identifier', v3_prefixes.identifier),
            ('file-name', v3_prefixes.file_name),
        ])
        code_gen_node[prefix_prop_name] = v3_prefix_node

        if v2_options_node is not None:
            header_node = collections.OrderedDict()
            _copy_prop_if_exists(header_node, v2_options_node, 'gen-prefix-def',
                                 'identifier-prefix-definition')
            _copy_prop_if_exists(header_node, v2_options_node, 'gen-default-stream-def',
                                 'default-data-stream-type-name-definition')
            code_gen_node['header'] = header_node

        self._root_node[opt_prop_name] = collections.OrderedDict({
            'code-generation': code_gen_node,
        })

        # convert the v2 metadata node into a v3 trace node
        try:
            self._root_node['trace'] = self._conv_meta_node(self._root_node['metadata'])
        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, 'Metadata object')

        del self._root_node['metadata']

    # Expands the field type aliases found in the metadata node.
    #
    # This method modifies the metadata node.
    #
    # When this method returns:
    #
    # * Any field type alias is replaced with its full field type node
    #   equivalent.
    #
    # * The `type-aliases` property of metadata node is removed.
    def _expand_ft_aliases(self):
        meta_node = self._root_node['metadata']
        ft_aliases_node = meta_node['type-aliases']

        # Expand field type aliases within trace, data stream, and event
        # record types now.
        try:
            self._resolve_ft_alias_from(ft_aliases_node, meta_node['trace'], 'packet-header-type')
        except _ConfigurationParseError as exc:
            _append_error_ctx(exc, 'Trace type')

        for dst_name, dst_node in meta_node['streams'].items():
            try:
                self._resolve_ft_alias_from(ft_aliases_node, dst_node, 'packet-context-type')
                self._resolve_ft_alias_from(ft_aliases_node, dst_node, 'event-header-type')
                self._resolve_ft_alias_from(ft_aliases_node, dst_node, 'event-context-type')

                for ert_name, ert_node in dst_node['events'].items():
                    try:
                        self._resolve_ft_alias_from(ft_aliases_node, ert_node, 'context-type')
                        self._resolve_ft_alias_from(ft_aliases_node, ert_node, 'payload-type')
                    except _ConfigurationParseError as exc:
                        _append_error_ctx(exc, f'Event record type `{ert_name}`')
            except _ConfigurationParseError as exc:
                _append_error_ctx(exc, f'Data stream type `{dst_name}`')

        # remove the (now unneeded) `type-aliases` node
        del meta_node['type-aliases']

    # Applies field type inheritance to all field type nodes found in
    # the metadata node.
    #
    # This method modifies the metadata node.
    #
    # When this method returns, no field type node has an `$inherit` or
    # `inherit` property.
    def _apply_fts_inheritance(self):
        meta_node = self._root_node['metadata']
        self._apply_ft_inheritance(meta_node['trace'], 'packet-header-type')

        for dst_node in meta_node['streams'].values():
            self._apply_ft_inheritance(dst_node, 'packet-context-type')
            self._apply_ft_inheritance(dst_node, 'event-header-type')
            self._apply_ft_inheritance(dst_node, 'event-context-type')

            for ert_node in dst_node['events'].values():
                self._apply_ft_inheritance(ert_node, 'context-type')
                self._apply_ft_inheritance(ert_node, 'payload-type')

    # Calls _expand_ft_aliases() and _apply_fts_inheritance() if the
    # metadata node has a `type-aliases` property.
    def _expand_fts(self):
        # Make sure that the current configuration node is valid
        # considering field types are not expanded yet.
        self._schema_validator.validate(self._root_node,
                                        'config/2/config-pre-field-type-expansion')

        meta_node = self._root_node['metadata']
        ft_aliases_node = meta_node.get('type-aliases')

        if ft_aliases_node is None:
            # If there's no `type-aliases` node, then there's no field
            # type aliases and therefore no possible inheritance.
            return

        # first, expand field type aliases
        self._expand_ft_aliases()

        # next, apply inheritance to create effective field types
        self._apply_fts_inheritance()

    # Processes the inclusions of the event record type node `ert_node`,
    # returning the effective node.
    def _process_ert_node_include(self, ert_node: _MapNode) -> _MapNode:
        # Make sure the event record type node is valid for the
        # inclusion processing stage.
        self._schema_validator.validate(ert_node, 'config/2/ert-pre-include')

        # process inclusions
        return self._process_node_include(ert_node, self._process_ert_node_include)

    # Processes the inclusions of the data stream type node `dst_node`,
    # returning the effective node.
    def _process_dst_node_include(self, dst_node: _MapNode) -> _MapNode:
        def process_children_include(dst_node):
            prop_name = 'events'

            if prop_name in dst_node:
                erts_node = dst_node[prop_name]

                for key in list(erts_node):
                    erts_node[key] = self._process_ert_node_include(erts_node[key])

        # Make sure the data stream type node is valid for the inclusion
        # processing stage.
        self._schema_validator.validate(dst_node, 'config/2/dst-pre-include')

        # process inclusions
        return self._process_node_include(dst_node, self._process_dst_node_include,
                                          process_children_include)

    # Processes the inclusions of the trace type node `trace_type_node`,
    # returning the effective node.
    def _process_trace_type_node_include(self, trace_type_node: _MapNode) -> _MapNode:
        # Make sure the trace type node is valid for the inclusion
        # processing stage.
        self._schema_validator.validate(trace_type_node, 'config/2/trace-type-pre-include')

        # process inclusions
        return self._process_node_include(trace_type_node, self._process_trace_type_node_include)

    # Processes the inclusions of the clock type node `clk_type_node`,
    # returning the effective node.
    def _process_clk_type_node_include(self, clk_type_node: _MapNode) -> _MapNode:
        # Make sure the clock type node is valid for the inclusion
        # processing stage.
        self._schema_validator.validate(clk_type_node, 'config/2/clock-type-pre-include')

        # process inclusions
        return self._process_node_include(clk_type_node, self._process_clk_type_node_include)

    # Processes the inclusions of the metadata node `meta_node`,
    # returning the effective node.
    def _process_meta_node_include(self, meta_node: _MapNode) -> _MapNode:
        def process_children_include(meta_node: _MapNode):
            prop_name = 'trace'

            if prop_name in meta_node:
                meta_node[prop_name] = self._process_trace_type_node_include(meta_node[prop_name])

            prop_name = 'clocks'

            if prop_name in meta_node:
                clk_types_node = meta_node[prop_name]

                for key in list(clk_types_node):
                    clk_types_node[key] = self._process_clk_type_node_include(clk_types_node[key])

            prop_name = 'streams'

            if prop_name in meta_node:
                dsts_node = meta_node[prop_name]

                for key in list(dsts_node):
                    dsts_node[key] = self._process_dst_node_include(dsts_node[key])

        # Make sure the metadata node is valid for the inclusion
        # processing stage.
        self._schema_validator.validate(meta_node, 'config/2/metadata-pre-include')

        # process inclusions
        return self._process_node_include(meta_node, self._process_meta_node_include,
                                          process_children_include)

    # Processes the inclusions of the configuration node, modifying it
    # during the process.
    def _process_config_includes(self):
        # Process inclusions in this order:
        #
        # 1. Clock type node, event record type nodes, and trace type
        #    nodes (the order between those is not important).
        #
        # 2. Data stream type nodes.
        #
        # 3. Metadata node.
        #
        # This is because:
        #
        # * A metadata node can include clock type nodes, a trace type
        #   node, data stream type nodes, and event record type nodes
        #   (indirectly).
        #
        # * A data stream type node can include event record type nodes.
        #
        # First, make sure the configuration node itself is valid for
        # the inclusion processing stage.
        self._schema_validator.validate(self._root_node,
                                        'config/2/config-pre-include')

        # Process metadata node inclusions.
        #
        # self._process_meta_node_include() returns a new (or the same)
        # metadata node without any `$include` property in it,
        # recursively.
        prop_name = 'metadata'
        self._root_node[prop_name] = self._process_meta_node_include(self._root_node[prop_name])

    def _parse(self):
        # Make sure the configuration node is minimally valid, that is,
        # it contains a valid `version` property.
        #
        # This step does not validate the whole configuration node yet
        # because we don't have an effective configuration node; we
        # still need to:
        #
        # * Process inclusions.
        # * Expand field types (aliases and inheritance).
        self._schema_validator.validate(self._root_node, 'config/2/config-min')

        # process configuration node inclusions
        self._process_config_includes()

        # Expand field type nodes.
        #
        # This process:
        #
        # 1. Replaces field type aliases with "effective" field type
        #    nodes, recursively.
        #
        #    After this step, the `type-aliases` property of the
        #    metadata node is gone.
        #
        # 2. Applies inheritance, following the `$inherit`/`inherit`
        #    properties.
        #
        #    After this step, field type nodes do not contain `$inherit`
        #    or `inherit` properties.
        #
        # This is done blindly, in that the process _doesn't_ validate
        # field type nodes at this point.
        #
        # The reason we must do this here for a barectf 2 configuration,
        # considering that barectf 3 also supports field type node
        # aliases and inheritance, is that we need to find specific
        # packet header and packet context field type member nodes (for
        # example, `stream_id`, `packet_size`, or `timestamp_end`) to
        # set the `$features` properties of barectf 3 trace type and
        # data stream type nodes. Those field type nodes can be aliases,
        # contain aliases, or inherit from other nodes.
        self._expand_fts()

        # Validate the whole, (almost) effective configuration node.
        #
        # It's almost effective because the `log-level` property of
        # event record type nodes can be log level aliases. Log level
        # aliases are also a feature of a barectf 3 configuration node,
        # therefore this is compatible.
        self._schema_validator.validate(self._root_node, 'config/2/config')

        # Transform the current configuration node into a valid v3
        # configuration node.
        self._transform_config_node()

    @property
    def config_node(self) -> config_parse_common._ConfigNodeV3:
        return config_parse_common._ConfigNodeV3(typing.cast(_MapNode, self._root_node))



# The MIT License (MIT)
#
# Copyright (c) 2014-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import barectf.config_parse_common as barectf_config_parse_common
import barectf.version as barectf_version
import barectf.config as barectf_config
import barectf.config_file as barectf_config_file
import barectf.codegen as barectf_codegen
import barectf.typing as barectf_typing


# version API
__major_version__ = barectf_version.__major_version__
__minor_version__ = barectf_version.__minor_version__
__patch_version__ = barectf_version.__patch_version__
__pre_version__ = barectf_version.__pre_version__
__version__ = barectf_version.__version__


# common typing API
Index = barectf_typing.Index
Count = barectf_typing.Count
Id = barectf_typing.Id
Alignment = barectf_typing.Alignment
VersionNumber = barectf_typing.VersionNumber


# configuration API
_ArrayFieldType = barectf_config._ArrayFieldType
_BitArrayFieldType = barectf_config._BitArrayFieldType
_ConfigurationParseError = barectf_config_parse_common._ConfigurationParseError
_EnumerationFieldType = barectf_config._EnumerationFieldType
_FieldType = barectf_config._FieldType
_IntegerFieldType = barectf_config._IntegerFieldType
ByteOrder = barectf_config.ByteOrder
ClockType = barectf_config.ClockType
ClockTypeCTypes = barectf_config.ClockTypeCTypes
ClockTypeOffset = barectf_config.ClockTypeOffset
Configuration = barectf_config.Configuration
ConfigurationCodeGenerationHeaderOptions = barectf_config.ConfigurationCodeGenerationHeaderOptions
ConfigurationCodeGenerationOptions = barectf_config.ConfigurationCodeGenerationOptions
ConfigurationOptions = barectf_config.ConfigurationOptions
DEFAULT_FIELD_TYPE = barectf_config.DEFAULT_FIELD_TYPE
DisplayBase = barectf_config.DisplayBase
DynamicArrayFieldType = barectf_config.DynamicArrayFieldType
EnumerationFieldTypeMapping = barectf_config.EnumerationFieldTypeMapping
EnumerationFieldTypeMappingRange = barectf_config.EnumerationFieldTypeMappingRange
EnumerationFieldTypeMappings = barectf_config.EnumerationFieldTypeMappings
EventRecordType = barectf_config.EventRecordType
LogLevel = barectf_config.LogLevel
RealFieldType = barectf_config.RealFieldType
SignedEnumerationFieldType = barectf_config.SignedEnumerationFieldType
SignedIntegerFieldType = barectf_config.SignedIntegerFieldType
StaticArrayFieldType = barectf_config.StaticArrayFieldType
DataStreamType = barectf_config.DataStreamType
DataStreamTypeEventRecordFeatures = barectf_config.DataStreamTypeEventRecordFeatures
DataStreamTypeFeatures = barectf_config.DataStreamTypeFeatures
DataStreamTypePacketFeatures = barectf_config.DataStreamTypePacketFeatures
StringFieldType = barectf_config.StringFieldType
StructureFieldType = barectf_config.StructureFieldType
StructureFieldTypeMember = barectf_config.StructureFieldTypeMember
StructureFieldTypeMembers = barectf_config.StructureFieldTypeMembers
Trace = barectf_config.Trace
TraceEnvironment = barectf_config.TraceEnvironment
TraceType = barectf_config.TraceType
TraceTypeWithUnknownNativeByteOrder = barectf_config.TraceTypeWithUnknownNativeByteOrder
TraceTypeFeatures = barectf_config.TraceTypeFeatures
UnsignedEnumerationFieldType = barectf_config.UnsignedEnumerationFieldType
UnsignedIntegerFieldType = barectf_config.UnsignedIntegerFieldType


# configuration file API
configuration_file_major_version = barectf_config_file.configuration_file_major_version
configuration_from_file = barectf_config_file.configuration_from_file
effective_configuration_file = barectf_config_file.effective_configuration_file


# code generation API
CodeGenerator = barectf_codegen.CodeGenerator


# remove local names
del barectf_config_parse_common
del barectf_version
del barectf_config
del barectf_config_file
del barectf_codegen
del barectf_typing



# The MIT License (MIT)
#
# Copyright (c) 2015-2020 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import barectf.config_parse as barectf_config_parse
import barectf.config as barectf_config
from barectf.typing import Count, VersionNumber
from typing import Optional, List, TextIO


def effective_configuration_file(file: TextIO, with_package_inclusion_directory: bool = True,
                                 inclusion_directories: Optional[List[str]] = None,
                                 ignore_inclusion_not_found: bool = False,
                                 indent_space_count: Count = Count(2)) -> str:
    if inclusion_directories is None:
        inclusion_directories = []

    return barectf_config_parse._effective_config_file(file, with_package_inclusion_directory,
                                                       inclusion_directories,
                                                       ignore_inclusion_not_found,
                                                       indent_space_count)


def configuration_from_file(file: TextIO, with_package_inclusion_directory: bool = True,
                            inclusion_directories: Optional[List[str]] = None,
                            ignore_inclusion_not_found: bool = False) -> barectf_config.Configuration:
    if inclusion_directories is None:
        inclusion_directories = []

    return barectf_config_parse._from_file(file, with_package_inclusion_directory,
                                           inclusion_directories, ignore_inclusion_not_found)


def configuration_file_major_version(file: TextIO) -> VersionNumber:
    return barectf_config_parse._config_file_major_version(file)
